from __future__ import annotations

import functools
from collections.abc import Callable, Collection
from typing import cast

import numpy as np
from navground import sim

from ..config import GroupConfig
from ..internal.group import create_agents_in_groups
from ..types import Reward


class RewardProbe(sim.RecordProbe):
    """
    A probe to record rewards to a single dataset
    of shape ``(#agents, #steps)``

    :param      ds:      The dataset
    :param      groups:  The configuration of the groups
    :param      reward:  The reward function

    """

    dtype = np.float64

    def __init__(self,
                 ds: sim.Dataset,
                 groups: Collection[GroupConfig] = tuple(),
                 reward: Reward | None = None):
        super().__init__(ds)
        self._groups = groups
        self._reward: dict[int, Reward] = {}
        self._default_reward = reward

    def prepare(self, run: sim.ExperimentalRun) -> None:
        agents = create_agents_in_groups(run.world, self._groups)
        for index, agent in agents.items():
            if agent.reward is not None:
                self._reward[index] = agent.reward
        super().prepare(run)

    def update(self, run: sim.ExperimentalRun) -> None:
        for i, agent in enumerate(run.world.agents):
            reward = self._reward.get(i, self._default_reward)
            if reward:
                value = reward(agent=agent,
                               world=run.world,
                               time_step=run.time_step)
                self.data.push(value)

    def get_shape(self, world: sim.World) -> list[int]:
        if self._default_reward:
            return [len(world.agents)]
        return [len(self._reward)]

    @classmethod
    def with_reward(cls,
                    reward: Reward) -> Callable[[sim.Dataset], RewardProbe]:
        """
        Creates a probe factory to record a reward

        :param      reward:  The reward

        :returns:   A callable that can be added to runs or experiments using
            :py:meth:`navground.sim.ExperimentalRun.add_record_probe` or
            :py:meth:`navground.sim.Experiment.add_record_probe`.
        """
        probe = functools.partial(cls, reward=reward)
        return cast('Callable[[sim.Dataset], RewardProbe]', probe)



from .evaluate_policies import evaluate_policies
from .evaluate_policy import evaluate_policy
from .experiment import make_experiment, make_experiment_with_env
from .logging import TrajectoryPlotConfig, VideoConfig, config_eval_log
from .navground import (evaluate, evaluate_with_experiment,
                        evaluate_with_experiment_and_env)
from .scenario import InitPolicyBehavior
from .video import (display_episode_video, display_run_video,
                    record_episode_video, record_run_video)

__all__ = [
    "evaluate_policy",
    "evaluate_policies",
    "InitPolicyBehavior",
    "make_experiment",
    "make_experiment_with_env",
    "evaluate",
    "evaluate_with_experiment",
    "evaluate_with_experiment_and_env",
    "config_eval_log",
    "TrajectoryPlotConfig",
    "VideoConfig",
    "display_run_video",
    "record_run_video",
    "display_episode_video",
    "record_episode_video",
]



from __future__ import annotations

from navground import core, sim


class CorridorWithObstacle(sim.Scenario,
                           name="CorridorWithObstacle"):
    """
    Simple worlds with

    - one agent that starts at ``(0, ~U(0, width))``
      and wants to travel towards ``+x``
    - two long horizontal walls at ``y = 0`` and ``y = width``
    - one circular obstacle at ``~(length, U(0, width))`` with radius
      ``~U(min_radius, max_radius)``

    The simulation finishes when the agent reaches ``x >= 2 length``

    :param      length:      The position of the obstacle
    :param      width:       The width of the corridor
    :param      min_radius:  The minimal obstacle radius
    :param      max_radius:  The maximal obstacle radius

    """
    _length = 10.0
    _width = 1.0
    _min_radius = 0.1
    _max_radius = 0.5

    def __init__(self,
                 length: float = 10.0,
                 width: float = 1.0,
                 min_radius: float = 0.1,
                 max_radius: float = 0.5):
        super().__init__()
        self._length = length
        self._width = width
        self._min_radius = min_radius
        self._max_radius = max_radius

    def init_world(self, world: sim.World, seed: int | None = None) -> None:
        """
        Initializes the world.

        :param      world:  The world
        :param      seed:   The seed

        """
        super().init_world(world, seed=seed)
        world.bounding_box = sim.BoundingBox(0, 2 * self._length, 0,
                                             self._width)
        world.add_wall(
            core.LineSegment((-self._length, 0), (2 * self._length, 0)))
        world.add_wall(
            core.LineSegment((-self._length, self._width),
                             (3 * self._length, self._width)))

        assert len(world.agents) == 1
        rng = world.random_generator
        agent = world.agents[0]
        r = agent.radius
        if agent.behavior:
            r += agent.behavior.safety_margin
            agent.twist = core.Twist2((agent.behavior.optimal_speed, 0), 0)
        agent.pose = core.Pose2((0, rng.uniform(r, self._width - r)), 0)
        agent.controller.follow_direction((1, 0))
        agent.last_cmd = agent.twist

        def t(world: sim.World) -> bool:
            return bool(agent.position[0] > 2 * self._length)

        world.set_termination_condition(t)

        ro = rng.uniform(self._min_radius, self._max_radius)
        margin = ro + r
        # TODO(Jerome): should be 2
        if rng.integers(1):
            y = rng.uniform(ro, self._width - margin)
        else:
            y = rng.uniform(margin, self._width - ro)
        world.add_obstacle(core.Disc((self._length, y), ro))

    @property
    @sim.register(10.0, "length of the corridor")
    def length(self) -> float:
        """
        The length of the corridor as the position of the obstacle
        """
        return self._length

    @length.setter
    def length(self, value: float) -> None:
        self._length = max(0, value)

    @property
    @sim.register(1.0, "width of the corridor")
    def width(self) -> float:
        """The width of the corridor"""
        return self._width

    @width.setter
    def width(self, value: float) -> None:
        self._width = max(0, value)

    @property
    @sim.register(0.2, "min radius")
    def min_radius(self) -> float:
        """The minimal obstacle radius"""
        return self._min_radius

    @min_radius.setter
    def min_radius(self, value: float) -> None:
        self._min_radius = max(0, value)

    @property
    @sim.register(0.2, "max radius")
    def max_radius(self) -> float:
        """The maximal obstacle radius"""
        return self._max_radius

    @max_radius.setter
    def max_radius(self, value: float) -> None:
        self._max_radius = max(0, value)



from __future__ import annotations

import sys
from collections.abc import Iterable, Sequence
from typing import TYPE_CHECKING, Any, NamedTuple, cast

import gymnasium as gym
import numpy as np
from matplotlib import pyplot as plt

from ..types import AnyPolicyPredictor

if TYPE_CHECKING:
    from matplotlib.axes import Axes
    from pandas import DataFrame


class LogField(NamedTuple):
    key: str
    label: str
    low: float | None = None
    high: float | None = None
    linestyle: str = '-'
    color: str | None = None


def plot_logs(logs: DataFrame,
              key: str,
              fields: Sequence[LogField],
              two_axis: bool = False,
              title: str = '',
              **kwargs: Any) -> None:
    """
    Plots logged fields.

    :param      logs:     The logs
    :param      key:      Common x-axis key
    :param      fields:   Which fields to plot.
    :param      two_axis: Whether to use two axis (only if there are two fields)

    """
    ts = logs[key]
    if two_axis and len(fields) != 2:
        print("two_axis=True only valid when plotting 2 fields!",
              file=sys.stderr)
        two_axis = False
    if two_axis:
        fig, ax = plt.subplots(**kwargs)
        for i, field in enumerate(fields):
            vs = logs[field.key]
            if i == 1:
                ax = ax.twinx()
            ax.plot(ts,
                    vs,
                    color=field.color,
                    linestyle=field.linestyle,
                    label=field.label)
            if i == 0:
                ax.legend(bbox_to_anchor=(0., 1.02, 1., .102),
                          loc='lower left')
                ax.set_xlabel('time steps')
            else:
                ax.legend(bbox_to_anchor=(0., 1.02, 1., .102),
                          loc='lower right')
            ax.set_ylabel(field.label)
        if title:
            plt.title(title)
    else:
        _ = plt.figure(**kwargs)
        scaled = False
        for field in fields:
            vs = logs[field.key]
            if field.low is not None and field.high is not None:
                vs = (vs - field.low) / (field.high - field.low)
                label = f'{field.label} (scaled)'
                scaled = True
            else:
                label = field.label
            plt.plot(ts,
                     vs,
                     color=field.color,
                     linestyle=field.linestyle,
                     label=label)
        if scaled:
            plt.ylim(-0.1, 1.1)
        plt.xlabel('time steps')
        if title:
            plt.title(title)
        plt.legend()


def plot_policy(policy: AnyPolicyPredictor,
                axs: Iterable[Axes] | None = None,
                fix: dict[str, float | Iterable[float]] = {},
                colors: Iterable | None = None,
                variable: dict[str, tuple[float, float]] = {},
                actions: dict[int, str] = {},
                cmap: str = 'RdYlGn',
                samples: int = 101,
                width: float = 5,
                height: float = 3,
                label: str = '',
                **kwargs: Any):
    in_space = policy.observation_space
    out_space = policy.action_space
    if not isinstance(in_space, gym.spaces.Dict):
        raise NotImplementedError
    variable_space = {k: v for k, v in in_space.items() if k not in fix}
    in_num = len(variable_space)
    if in_num > 2:
        raise ValueError(f"Too many variables: {', '.join(variable_space)}!")
    if in_num == 0:
        raise ValueError("No variable")
    for k, s in variable_space.items():
        space = cast('gym.spaces.Box', s)
        if k not in variable:
            variable[k] = (space.low.flatten()[0], space.high.flatten()[0])
    out_num = len(actions)
    for k, v in fix.items():
        if isinstance(v, Iterable):
            vs = list(v)
            if len(vs) == 0:
                raise ValueError(f"Empty sequence for fixed {k}")
            elif len(vs) == 1:
                fix[k] = vs[0]
                continue
            else:
                if in_num == 2:
                    if not axs:
                        figsize = (width * out_num, height * len(vs))
                        _, all_axs = plt.subplots(ncols=out_num,
                                                  nrows=len(vs),
                                                  figsize=figsize)
                        all_axs = np.atleast_2d(all_axs)
                    else:
                        all_axs = axs
                    for f_v, f_axs in zip(vs, all_axs.T, strict=True):
                        f_fix = dict(**fix)
                        f_fix[k] = f_v
                        plot_policy(policy=policy,
                                    axs=f_axs,
                                    fix=f_fix,
                                    variable=variable,
                                    actions=actions,
                                    cmap=cmap,
                                    samples=samples,
                                    label=f'{k}={round(f_v, 2)}')
                    return
                else:
                    if axs is None:
                        figsize = (width * out_num, height)
                        _, all_axs = plt.subplots(ncols=out_num,
                                                  nrows=1,
                                                  figsize=figsize)
                        all_axs = np.atleast_1d(all_axs)
                    else:
                        all_axs = axs
                    if colors is None:
                        colors = [None] * len(vs)
                    for f_v, color in zip(vs, colors, strict=True):
                        f_fix = dict(**fix)
                        f_fix[k] = f_v
                        plot_policy(policy,
                                    axs=all_axs,
                                    variable=variable,
                                    actions=actions,
                                    cmap=cmap,
                                    samples=samples,
                                    fix=f_fix,
                                    label=f'{k}={round(f_v, 2)}',
                                    color=color)
                    for ax in all_axs:
                        ax.legend()
                    return
    if axs is None:
        figsize = (width * out_num, height)
        _, all_axs = plt.subplots(ncols=out_num, nrows=1, figsize=figsize)
        axs = np.atleast_1d(all_axs)
    obs = {}
    variables = {
        k: np.linspace(low, high, samples, dtype=np.float32)
        for k, (low, high) in variable.items()
    }
    if in_num == 2:
        xs, ys = list(variables.values())
        obs = {
            k: v
            for k, v in zip(variables, np.meshgrid(xs, ys), strict=True)
        }
        size = samples**2
    else:
        obs = dict(**variables)
        size = samples
    for k, v in fix.items():
        obs[k] = np.full(size, v, dtype=np.float32)
    obs = {
        k: obs[k].reshape(-1,
                          *cast('gym.spaces.Box', v).shape)
        for k, v in in_space.items()
    }
    act, _ = policy.predict(obs, deterministic=True)
    if in_num == 2:
        act = act.reshape(samples, samples, -1)
    if isinstance(out_space, gym.spaces.Discrete):
        act = act - 1
    elif isinstance(out_space, gym.spaces.MultiBinary):
        shape = act.shape
        act = act.reshape(*shape[:-1], -1, 2)
        act = act[..., 1] - act[..., 0]
    if in_num == 2:
        (xlabel, xs), (ylabel, ys) = list(variables.items())

        for ax, (i, out_label) in zip(axs, actions.items(), strict=True):
            low = -1
            high = 1
            im = ax.imshow(act[::-1, :, i],
                           vmin=low,
                           vmax=high,
                           cmap=cmap,
                           **kwargs)
            ax.set_xlabel(xlabel)
            ax.set_ylabel(ylabel)
            plt.colorbar(im, ax=ax)
            ps = np.linspace(0, samples, 5)
            xs = np.linspace(xs[0], xs[-1], 5)
            ys = np.linspace(ys[0], ys[-1], 5)
            ax.set_xticks(ps, [f'{x:.2f}' for x in xs])
            ax.set_yticks(ps[::-1], [f'{y:.2f}' for y in ys])
            if label:
                title = f'{out_label} ({label})'
            else:
                title = out_label
            ax.set_title(title)
    else:
        xlabel, xs = next(iter(variables.items()))
        for ax, (i, out_label) in zip(axs, actions.items(), strict=True):
            ax.plot(xs, act[..., i], label=label, **kwargs)
            ax.set_xlabel(xlabel)
            ax.set_ylabel(out_label)
            ax.set_title(out_label)



from __future__ import annotations

from pathlib import Path

from stable_baselines3.common.callbacks import BaseCallback

from ...onnx import export


class ExportOnnxCallback(BaseCallback):
    """
    Exports the (best) model policy as "best_policy.onnx".
    """

    def _on_step(self) -> bool:
        p = self.model.logger.get_dir()
        if p is not None:
            path = Path(p) / 'best_policy.onnx'
            export(self.model.policy, path)
        return True



from __future__ import annotations

import warnings

from navground import core

from ...config import ControlActionConfig, DefaultObservationConfig
from ...types import AnyPolicyPredictor, ObservationTransform, PathLike
from .base import BasePolicyMixin


class PolicyBehavior(BasePolicyMixin, core.Behavior, name="Policy",
                     include_properties_of=[BasePolicyMixin]):
    """
    A navigation behavior that evaluates a ML policy

    *Registered properties*:

    - :py:attr:`policy_path` (str)
    - :py:attr:`flat` (bool)
    - :py:attr:`history` (int)
    - :py:attr:`fix_orientation` (bool)
    - :py:attr:`include_target_direction` (bool)
    - :py:attr:`include_target_direction_validity` (bool)
    - :py:attr:`include_target_distance` (bool)
    - :py:attr:`include_target_distance_validity` (bool)
    - :py:attr:`include_target_speed` (bool)
    - :py:attr:`include_target_angular_speed` (bool)
    - :py:attr:`include_velocity` (bool)
    - :py:attr:`include_angular_speed` (bool)
    - :py:attr:`include_radius` (bool)
    - :py:attr:`use_wheels` (bool)
    - :py:attr:`use_acceleration_action` (bool)
    - :py:attr:`max_acceleration` (float)
    - :py:attr:`max_angular_acceleration` (float)
    - :py:attr:`deterministic` (bool)

    *State*: :py:class:`core.SensingState`

    :param kinematics: The agent kinematics
    :param radius: The agent radius
    :param policy: The policy
    :param action_config: Configures which actions the policy generates
    :param observation_config: Configures which observations the policy consumes
    :param deterministic: Whether the policy evaluation is deterministic
    :param pre: Optional input (observations) transformation
    """

    def __init__(self,
                 kinematics: core.Kinematics | None = None,
                 radius: float = 0.0,
                 policy: AnyPolicyPredictor | None = None,
                 policy_path: PathLike = '',
                 action_config: ControlActionConfig = ControlActionConfig(),
                 observation_config:
                 DefaultObservationConfig = DefaultObservationConfig(),
                 deterministic: bool = False,
                 pre: ObservationTransform | None = None):
        core.Behavior.__init__(self, kinematics, radius)
        BasePolicyMixin.__init__(self, policy, policy_path, action_config,
                                 observation_config, deterministic, pre)

    def prepare(self) -> None:
        BasePolicyMixin.prepare(self)
        if not self._policy:
            self.init_policy()

    def compute_cmd_internal(self, time_step: float) -> core.Twist2:
        # if self.check_if_target_satisfied():
        #     return core.Twist2((0, 0), 0, frame=core.Frame.relative)
        self.prepare()
        if not self._policy:
            warnings.warn("Policy not set", stacklevel=1)
            return core.Twist2((0, 0), 0, frame=core.Frame.relative)
        obs = self.update_observation()
        if self._pre:
            obs = self._pre(obs)
        act, _ = self._policy.predict(obs, deterministic=self.deterministic)
        return self.get_cmd_from_action(act, time_step)



from __future__ import annotations

import inspect
import os
from collections.abc import Callable, Mapping, Sequence
from typing import TYPE_CHECKING, Any, Protocol, TypeAlias

if TYPE_CHECKING:
    import gymnasium as gym
    import torch

try:
    from typing import Self
except ImportError:
    try:
        from typing_extensions import Self
    except ImportError:
        pass
import abc

import numpy as np
from navground import sim

from .register import Registrable

Array: TypeAlias = np.typing.NDArray[Any]
Observation: TypeAlias = dict[str, Array] | Array
Action: TypeAlias = Array
State: TypeAlias = tuple[Array, ...]
EpisodeStart: TypeAlias = Array
Info: TypeAlias = list[dict[str, Array]] | dict[str, Array]
PathLike: TypeAlias = os.PathLike[str] | str

Bounds: TypeAlias = tuple[np.typing.NDArray[np.floating[Any]],
                          np.typing.NDArray[np.floating[Any]]]

ObservationTransform: TypeAlias = Callable[[Observation], Observation]

SensorLike: TypeAlias = sim.Sensor | str | dict[str, Any]
SensorSequenceLike: TypeAlias = Sequence[SensorLike] | str

TerminationCondition = Callable[[sim.Agent, sim.World], bool]


class JSONAble(Protocol):

    @property
    def asdict(self) -> dict[str, Any]:
        """
        Return a JSON-able representation

        :returns:  A JSON-able dict
        """
        ...

    @classmethod
    def from_dict(cls, value: Mapping[str, Any]) -> Self:
        """
        Load the class from the JSON representation

        :param value:  A JSON-able dict
        :returns: An instance of the class
        """
        ...


class Reward(abc.ABC, Registrable):
    """
    The reward protocol is a callable to compute
    scalar rewards for individual agents.
    """

    @abc.abstractmethod
    def __call__(self, agent: sim.Agent, world: sim.World,
                 time_step: float) -> float:
        """
        Compute the reward for an agent.

        :param      agent:      The agent
        :param      world:      The simulated world the agent belongs to
        :param      time_step:  The time step of the simulation

        :returns:   A scalar reward number
        """


class PolicyPredictor(Protocol):
    """
    This class describes the predictor protocol.

    Same as :py:type:`stable_baselines3.common.type_aliases.PolicyPredictor`,
    included here to be self-contained.
    """

    @property
    def action_space(self) -> gym.Space:
        ...

    @property
    def observation_space(self) -> gym.Space:
        ...

    def predict(self,
                observation: Observation,
                state: State | None = None,
                episode_start: EpisodeStart | None = None,
                deterministic: bool = False) -> tuple[Action, State | None]:
        """
        Get the policy action from an observation (and optional hidden state).
        Includes sugar-coating to handle different observations (e.g. normalizing images).

        :param observation: the input observation
        :param state: The last hidden states (can be None, used in recurrent policies)
        :param episode_start: The last masks (can be None, used in recurrent policies)
            this correspond to beginning of episodes,
            where the hidden states of the RNN must be reset.
        :param deterministic: Whether or not to return deterministic actions.
        :return: the model's action and the next hidden state
            (used in recurrent policies)
        """
        ...


class PolicyPredictorWithInfo(Protocol):
    """
    Similar to :py:class:`PolicyPredictor` but :py:meth:`predict`
    accepts info dictionaries.
    """

    @property
    def action_space(self) -> gym.Space:
        ...

    @property
    def observation_space(self) -> gym.Space:
        ...

    def predict(self,
                observation: Observation,
                state: State | None = None,
                episode_start: EpisodeStart | None = None,
                deterministic: bool = False,
                info: Info | None = None) -> tuple[Action, State | None]:
        """
        Get the policy action from an observation (and optional hidden state).
        Includes sugar-coating to handle different observations (e.g. normalizing images).

        :param observation: the input observation
        :param state: The last hidden states (can be None, used in recurrent policies)
        :param episode_start: The last masks (can be None, used in recurrent policies)
            this correspond to beginning of episodes,
            where the hidden states of the RNN must be reset.
        :param deterministic: Whether or not to return deterministic actions.
        :param info: Dictionaries with generic information that is not part of observation and state.
        :return: the model's action and the next hidden state
            (used in recurrent policies)
        """
        ...


AnyPolicyPredictor: TypeAlias = PolicyPredictor | PolicyPredictorWithInfo


def accept_info(func: Callable[..., Any]) -> bool:
    """
    Check whether the callable accept an ``info`` argument.

    Can be used to distinguish :py:class:`PolicyPredictorWithInfo`
    from :py:class:`PolicyPredictor`:

    >>> policy = MyPolicy(...)
    >>> accept_info(policy.predict)
    >>> False

    :param      func:  The function to test
    :returns:   True if it has an argument named ``info``
    """
    sig = inspect.signature(func)
    return 'info' in sig.parameters


PyTorchObs: TypeAlias = 'torch.Tensor | dict[str, torch.Tensor]'


class PyTorchPolicy(PolicyPredictor, Protocol):

    def __call__(self,
                 obs: PyTorchObs,
                 deterministic: bool = False) -> torch.Tensor:
        """
        Evaluate the policy

        :param      obs:            The observations
        :param      deterministic:  Whether or not to return deterministic actions.
        """
        ...

    def forward(self,
                obs: PyTorchObs,
                deterministic: bool = False) -> torch.Tensor:
        """
        Evaluate the policy

        :param      obs:            The observations
        :param      deterministic:  Whether or not to return deterministic actions.
        """
        ...



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from stable_baselines3.common.callbacks import EvalCallback

from ...types import PathLike
from .export_onnx_callback import ExportOnnxCallback
from .log import load_eval_logs, plot_eval_logs, plot_rollout_logs
from .progress_bar_with_reward import ProgressBarWithRewardCallback
from .video_callback import VideoCallback

if TYPE_CHECKING:
    from stable_baselines3.common.callbacks import BaseCallback
    from stable_baselines3.common.vec_env import VecEnv


def callbacks(venv: VecEnv,
              best_model_save_path: PathLike,
              eval_freq: int = 1000,
              n_eval_episodes: int = 30,
              number_of_videos: int = 1,
              update_bar_every: int = 1,
              video_policy: Any = None,
              video_env: Any = None,
              export_to_onnx: bool = False,
              **kwargs: Any) -> list[BaseCallback]:
    if number_of_videos > 0:
        video_cb = VideoCallback(video_env or venv,
                                 number=number_of_videos,
                                 factor=4,
                                 policy=video_policy,
                                 **kwargs)
    else:
        video_cb = None
    callback_on_new_best = ExportOnnxCallback() if export_to_onnx else None
    eval_cb = EvalCallback(venv,
                           best_model_save_path=str(best_model_save_path),
                           eval_freq=eval_freq,
                           deterministic=True,
                           render=False,
                           n_eval_episodes=n_eval_episodes,
                           verbose=0,
                           callback_after_eval=video_cb,
                           callback_on_new_best=callback_on_new_best)
    return [eval_cb, ProgressBarWithRewardCallback(every=update_bar_every)]


__all__ = [
    "ProgressBarWithRewardCallback", "ExportOnnxCallback", "load_eval_logs",
    "plot_eval_logs", "VideoCallback", 'callbacks', 'plot_rollout_logs'
]



from __future__ import annotations

import math
from collections.abc import Callable, Collection, Sequence
from typing import TypeAlias, cast

import gymnasium as gym
import numpy as np
import torch as th
import torch.nn as nn
from stable_baselines3.common.preprocessing import (get_flattened_obs_dim,
                                                    is_image_space)
from stable_baselines3.common.torch_layers import (BaseFeaturesExtractor,
                                                   NatureCNN, create_mlp)
from torch.masked import as_masked_tensor

Reduction: TypeAlias = Callable[[th.Tensor, int, bool], th.Tensor]

# TODO: make sure images are not grouped.


class OrderInvariantCombinedExtractor(BaseFeaturesExtractor):
    """
    A variation of SB3 :py:class:`stable_baselines3.common.torch_layers.CombinedExtractor`
    that applies a ordering invariant MLP feature extractor to a group of keys
    after optionally masking it.

    Same as the original ``CombinedExtractor``:
    Combined features extractor for Dict observation spaces.
    Builds a features extractor for each key of the space. Input from each space
    is fed through a separate submodule (CNN or MLP, depending on input shape),
    the output features are concatenated and fed through additional MLP network ("combined").

    :param observation_space: the observation space
    :param cnn_output_dim: Number of features to output from each CNN submodule(s).
                           Defaults to 256 to avoid exploding network sizes.
    :param normalized_image: Whether to assume that the image is already normalized
        or not (this disables dtype and bounds checks): when True, it only checks that
        the space is a Box and has 3 dimensions.
        Otherwise, it checks that it has expected dtype (uint8) and bounds (values in [0, 255]).

    :param order_invariant_keys: the keys to group together and process by an
                                 ordering invariant feature extractor
    :param replicated_keys:  additional keys to add to the ordering invariant groups,
                             replicating the values for each group
    :param filter_key: the key to use for masking to select items with positive values of this key
    :param removed_keys: keys removed from the observations before concatenating
                         with ordering invariant features
    :param net_arch: the ordering invariant MLP layers sizes
    :param activation_fn: the ordering invariant MLP activation function
        If not set, it defaults to ``torch.nn.ReLU``.
    :param reductions: A sequence of (order-invariant) modules.
        If not set, it defaults to ``[torch.sum]``.

    """

    def __init__(self,
                 observation_space: gym.spaces.Dict,
                 cnn_output_dim: int = 256,
                 normalized_image: bool = False,
                 order_invariant_keys: Collection[str] = [],
                 replicated_keys: Collection[str] = [],
                 filter_key: str = "",
                 removed_keys: Collection[str] = [],
                 net_arch: list[int] = [8],
                 activation_fn: type[nn.Module] | None = None,
                 reductions: Sequence[Reduction] | None = None):

        super().__init__(observation_space, features_dim=1)
        if activation_fn is None:
            activation_fn = nn.ReLU
        if reductions is None:
            reductions = [th.sum]
        # Same as CombinedExtractor
        extractors: dict[str, nn.Module] = {}
        order_invariant_lens: set[int] = set()
        total_concat_size = 0
        order_invariant_size = 0
        replicated_size = 0
        for key, subspace in observation_space.spaces.items():
            if is_image_space(subspace, normalized_image=normalized_image):
                extractors[key] = NatureCNN(subspace,
                                            features_dim=cnn_output_dim,
                                            normalized_image=normalized_image)
                total_concat_size += cnn_output_dim
            else:
                extractors[key] = nn.Flatten()
                if key in order_invariant_keys:
                    order_invariant_lens.add(
                        cast("gym.spaces.Box", subspace).shape[0])
                    order_invariant_size += get_flattened_obs_dim(subspace)
                elif key in replicated_keys:
                    replicated_size += get_flattened_obs_dim(subspace)
                elif key == filter_key:
                    order_invariant_lens.add(
                        cast("gym.spaces.Box", subspace).shape[0])
                elif key not in removed_keys:
                    total_concat_size += get_flattened_obs_dim(subspace)

        self.extractors = nn.ModuleDict({
            k: v
            for k, v in extractors.items() if k not in order_invariant_keys
            and k != filter_key and k not in removed_keys
        })
        self.order_invariant_extractors = nn.ModuleDict({
            k: v
            for k, v in extractors.items() if k in order_invariant_keys
        })
        self.replicated_extractors = nn.ModuleDict({
            k: v
            for k, v in extractors.items() if k in replicated_keys
        })
        self.filter_key = filter_key
        assert len(order_invariant_lens) == 1
        self._number = order_invariant_lens.pop()
        order_invariant_size = order_invariant_size // self._number + replicated_size
        self.net_out_dim = net_arch[-1] * len(reductions)
        self._features_dim = total_concat_size + self.net_out_dim
        mlp = create_mlp(order_invariant_size, 0, net_arch, activation_fn)
        self.nn = nn.Sequential(*mlp)
        self.reductions = reductions

    def forward(self, observations: dict[str, th.Tensor]) -> th.Tensor:
        encoded_tensor_list = []
        bs = set([int(v.shape[0]) for k, v in observations.items()])
        assert len(bs) == 1, bs
        batches = next(iter(bs))
        for key, extractor in self.extractors.items():
            encoded_tensor_list.append(extractor(observations[key]))
        if self.order_invariant_extractors:
            oit = [
                extractor(observations[key]).reshape(batches, self._number, -1)
                for key, extractor in self.order_invariant_extractors.items()
            ]
            rt = [
                extractor(observations[key]).repeat(batches, self._number, -1)
                for key, extractor in self.replicated_extractors.items()
            ]
            order_invariant_input = th.cat(oit + rt, dim=-1)
            if self.filter_key:
                mask = observations[self.filter_key].flatten() > 0
                order_invariant_input = order_invariant_input[mask]
            z = self.nn(order_invariant_input)
            if len(z):
                for reduction in self.reductions:
                    r = reduction(z, 1, False)
                    if reduction in (th.max, th.min):
                        r = r.values  # type: ignore
                    encoded_tensor_list.append(r)
            else:
                encoded_tensor_list.append(th.zeros(1, self.net_out_dim))
        return th.cat(encoded_tensor_list, dim=-1)


class OrderInvariantFlattenExtractor(BaseFeaturesExtractor):
    """

    Similar to :py:class:`OrderInvariantCombinedExtractor`
    but for flat observation spaces.

    :param observation_space: the observation space
    :param order_invariant_slices: the slices to group together and process by
                                   an ordering invariant feature extractor
    :param replicated_slices: additional slices to add to the ordering invariant groups,
                              replicating the values for each group
    :param filter_slice: the slice to use for masking to select items with
                         positive values for indices in this slice
    :param removed_slices: keys removed from the observations before concatenating
                           with ordering invariant features
    :param net_arch: the ordering invariant MLP layers sizes
    :param activation_fn: the ordering invariant MLP activation function
        If not set, it defaults to ``torch.nn.ReLU``.
    :param reductions: A sequence of (order-invariant) modules.
        If not set, it defaults to ``[torch.sum]``.
    :param use_masked_tensors: Whether to use masked tensors
    """

    def __init__(self,
                 observation_space: gym.spaces.Box,
                 order_invariant_slices: Collection[slice] = [],
                 replicated_slices: Collection[slice] = [],
                 filter_slice: slice | None = None,
                 removed_slices: Collection[slice] = [],
                 number: int = 0,
                 net_arch: list[int] = [8],
                 activation_fn: type[nn.Module] | None = None,
                 reductions: Sequence[Reduction] | None = None,
                 use_masked_tensors: bool = False):

        super().__init__(observation_space, features_dim=1)
        if activation_fn is None:
            activation_fn = nn.ReLU
        if reductions is None:
            reductions = [th.sum]
        self.flatten = nn.Flatten()
        input_size = math.prod(observation_space.shape)
        self._invariant_slices = order_invariant_slices
        self._filter_slice = filter_slice
        self._number = number
        size = sum(s.stop - s.start for s in order_invariant_slices)
        self._non_invariant_mask = np.ones(input_size, dtype=bool)
        for s in order_invariant_slices:
            self._non_invariant_mask[s] = False
        for s in removed_slices:
            self._non_invariant_mask[s] = False
        if filter_slice:
            self._non_invariant_mask[filter_slice] = False
        self._replicated_mask = np.zeros(input_size, dtype=bool)
        for s in replicated_slices:
            self._replicated_mask[s] = True
        self._should_replicate = any(self._replicated_mask)
        total_concat_size = sum(self._non_invariant_mask)
        order_invariant_size = size // number + sum(self._replicated_mask)
        self.net_out_dim = net_arch[-1] * len(reductions)
        self._features_dim = total_concat_size + self.net_out_dim
        mlp = create_mlp(order_invariant_size, 0, net_arch, activation_fn)
        self.nn = nn.Sequential(*mlp)
        self.reductions = reductions
        self.use_masked_tensors = use_masked_tensors

    def forward(self, observations: th.Tensor) -> th.Tensor:
        observations = self.flatten(observations)
        shape = observations.shape
        ni = observations[..., self._non_invariant_mask]
        if self._invariant_slices:
            if self._filter_slice:
                mask = (observations[..., self._filter_slice] > 0).detach()
                numbers = th.sum(mask, -1)
                has_same_number = len(th.unique(numbers)) == 1
            else:
                has_same_number = True
                mask = None

            ois = [
                observations[..., s].reshape(*shape[:-1], self._number, -1)
                for s in self._invariant_slices
            ]
            if self._should_replicate:
                ri = [1] * (len(shape) + 1)
                ri[-2] = self._number
                r = observations[..., self._replicated_mask].reshape(
                    *shape[:-1], 1, -1)
                ois.append(r.repeat(*ri))
            oi = th.cat(ois, dim=-1)
            if mask is not None and not self.use_masked_tensors:
                oi = oi[mask]
                if has_same_number:
                    number = cast("int", numbers[0].item())
                    if number > 0:
                        oi = oi.reshape(*shape[:-1], number, -1)

            zs = [ni]
            if len(oi):
                z = self.nn(oi)
                if self.use_masked_tensors and mask is not None:
                    ri = [1] * len(z.shape)
                    ri[-1] = z.shape[-1]
                    mask = mask.unsqueeze(-1).repeat(*ri)
                    z = as_masked_tensor(z,
                                         mask)  # type: ignore[no-untyped-call]
                    for reduction in self.reductions:
                        r = reduction(z, -2, False)
                        zs.append(r)
                        # TODO(Jerome): check after linting
                        # zs.append(r.to_tensor(0))
                elif has_same_number:
                    for reduction in self.reductions:
                        r = reduction(z, -2, False)
                        # TODO(Jerome): if we restrict to amax/amin, no need to
                        # check output
                        if isinstance(r, th.Tensor):
                            zs.append(r)
                        else:
                            zs.append(r.values)
                else:
                    i = 0
                    rs = []
                    for tn in numbers:
                        n = tn.item()
                        if n:
                            bs = []
                            mz = z[i:i + n]
                            for reduction in self.reductions:
                                r = reduction(mz, -2, False)
                                if isinstance(r, th.Tensor):
                                    bs.append(r)
                                else:
                                    bs.append(r.values)
                            rs.append(th.concat(bs))
                        else:
                            rs.append(th.zeros(self.net_out_dim))
                            # rs.extend([th.zeros(z.shape[1:])] * len(self.reductions))
                        i += n
                    zs.append(th.stack(rs))
            else:
                zs.append(th.zeros(*shape[:-1], self.net_out_dim))
            return th.cat(zs, dim=-1)
        return ni


def make_order_invariant_flatten_extractor(
        observation_space: gym.spaces.Box,
        dict_space: gym.spaces.Dict,
        cnn_output_dim: int = 256,
        normalized_image: bool = False,
        order_invariant_keys: Collection[str] = [],
        replicated_keys: Collection[str] = [],
        filter_key: str = "",
        removed_keys: Collection[str] = [],
        net_arch: list[int] = [8],
        activation_fn: type[nn.Module] | None = None,
        reductions: Sequence[Reduction] | None = None,
        use_masked_tensors: bool = False) -> OrderInvariantFlattenExtractor:
    """

    Helper function that creates a :py:class:`OrderInvariantFlattenExtractor`
    using information from a dictionary space to infer the layout of the
    observation space.

    :param observation_space: the observation space
    :param      dict_space:            The dictionary space
    :param cnn_output_dim: Number of features to output from each CNN submodule(s).
                           Defaults to 256 to avoid exploding network sizes.
    :param normalized_image: Whether to assume that the image is already normalized
        or not (this disables dtype and bounds checks): when True, it only checks that
        the space is a Box and has 3 dimensions.
        Otherwise, it checks that it has expected dtype (uint8) and bounds (values in [0, 255]).

    :param order_invariant_keys: the keys to group together and process by an
                                 ordering invariant feature extractor
    :param replicated_keys:  additional keys to add to the ordering invariant groups,
                             replicating the values for each group
    :param filter_key: the key to use for masking to select items with positive values of this key
    :param removed_keys: keys removed from the observations before concatenating
                         with ordering invariant features
    :param net_arch: the ordering invariant MLP layers sizes
    :param activation_fn: the ordering invariant MLP activation function
        If not set, it defaults to ``torch.nn.ReLU``.
    :param reductions: A sequence of (order-invariant) modules.
        If not set, it defaults to ``[torch.sum]``.
    :param use_masked_tensors:   Whether to use masked tensors

    :returns:   The order invariant flatten extractor.

    :raises     AssertionError:  if ``filter_key`` is associated
        to a space that is non-flat.
    """

    ns = set(
        cast("gym.spaces.Box", dict_space[k]).shape[0]
        for k in order_invariant_keys)
    if filter_key and filter_key in dict_space:
        ns.add(cast("gym.spaces.Box", dict_space[filter_key]).shape[0])
    assert len(ns) == 1
    number = list(ns)[0]
    indices = {}
    i = 0
    for k, v in dict_space.items():
        dim = gym.spaces.flatdim(v)
        indices[k] = slice(i, i + dim, 1)
        i += dim
    invariant_slices = [
        indices[k] for k in order_invariant_keys if k in indices
    ]
    removed_slices = [indices[k] for k in removed_keys if k in indices]
    replicated_slices = [indices[k] for k in replicated_keys if k in indices]
    filter_slice = indices.get(filter_key)
    return OrderInvariantFlattenExtractor(
        observation_space=observation_space,
        order_invariant_slices=invariant_slices,
        replicated_slices=replicated_slices,
        filter_slice=filter_slice,
        removed_slices=removed_slices,
        number=number,
        net_arch=net_arch,
        activation_fn=activation_fn,
        reductions=reductions,
        use_masked_tensors=use_masked_tensors)



from .env import (BaseParallelEnv, MultiAgentNavgroundEnv,
                  make_shared_parallel_env_with_env, parallel_env,
                  shared_parallel_env)
from .join import JointEnv
from .utils import make_vec_from_penv

__all__ = [
    'MultiAgentNavgroundEnv', 'BaseParallelEnv', 'JointEnv',
    'shared_parallel_env', 'parallel_env', 'make_vec_from_penv',
    'make_shared_parallel_env_with_env'
]



from __future__ import annotations

import time


class SyncClock:

    def __init__(self, time_step: float, factor: float = 1.0):
        self._last_time: float | None = None
        self._period = time_step / factor

    def tick(self) -> None:
        now = time.monotonic()
        if self._last_time is not None:
            sleep_time = max(self._last_time + self._period - now, 0.0)
            if sleep_time > 0:
                time.sleep(sleep_time)
                now = time.monotonic()
        self._last_time = now



from __future__ import annotations

from collections.abc import Callable, Collection
from typing import TYPE_CHECKING, Any

import numpy as np

from navground import sim
from navground.sim.ui.video import display_video, record_video

from ..config import GroupConfig
from .experiment import make_experiment_with_env

if TYPE_CHECKING:
    from stable_baselines3.common.vec_env import VecEnv

    from ..env import BaseEnv
    from ..parallel_env import BaseParallelEnv
    from ..types import AnyPolicyPredictor, PathLike


def _run_video(fn: Callable,
               experiment: sim.Experiment,
               factor: int = 1,
               seed: int = 0,
               use_world_bounds: bool = True,
               select: int = 0,
               of: int = 0,
               **kwargs) -> Any:
    if of:
        experiment.run_index = seed
        experiment.run(number_of_runs=of)
        rs = sorted(((np.min(np.sum(run.records['reward'], axis=0)), i)
                     for i, run in experiment.runs.items()),
                    reverse=True)
        index = rs[select][1]
        run = experiment.runs[index]
    else:
        run = experiment.run_once(seed)
    world = experiment.scenario.make_world(run.seed)
    if use_world_bounds:
        bounds = sim.bounds_for_world(world)
    else:
        bounds = None
    if run.final_sim_time:
        r = fn(world=world,
               time_step=run.time_step,
               duration=run.final_sim_time,
               factor=factor,
               fps=factor / run.time_step,
               bounds=bounds,
               **kwargs)
    else:
        r = None
    world._close()
    return r


def display_run_video(experiment: sim.Experiment,
                      factor: int = 1,
                      seed: int = 0,
                      use_world_bounds: bool = True,
                      select: int = 0,
                      of: int = 0,
                      **kwargs: Any) -> Any:
    """
    Display a video from one episode from an evaluation experiment.

    :param      experiment:        The experiment
    :param      factor:            The real-time factor
    :param      seed:              The seed (only applies if ``of == 0``)
    :param      use_world_bounds:  Whether to keep the initial world bounds
    :param      select:            Which episode to select
                                   (ordered by reward, only applies if ``of > 0``)
    :param      of:                The number of runs to choose the episode from
    :param      kwargs:            Keywords arguments passed to the renderer
    """
    return _run_video(display_video,
                      experiment,
                      factor=factor,
                      seed=seed,
                      use_world_bounds=use_world_bounds,
                      select=select,
                      of=of,
                      **kwargs)


def record_run_video(experiment: sim.Experiment,
                     path: PathLike,
                     factor: int = 1,
                     seed: int = 0,
                     use_world_bounds: bool = True,
                     select: int = 0,
                     of: int = 0,
                     **kwargs: Any) -> None:
    """
    Record a video from one episode from an evaluation experiment.

    :param      experiment:        The experiment
    :param      path:              Where to save the video
    :param      factor:            The real-time factor
    :param      seed:              The seed (only applies if ``of == 0``)
    :param      use_world_bounds:  Whether to keep the initial world bounds
    :param      select:            Which episode to select
                                   (ordered by reward, only applies if ``of > 0``)
    :param      of:                The number of runs to choose the episode from
    :param      kwargs:            Keywords arguments passed to the renderer

    """
    return _run_video(record_video,
                      experiment,
                      path=path,
                      factor=factor,
                      seed=seed,
                      use_world_bounds=use_world_bounds,
                      select=select,
                      of=of,
                      **kwargs)


def display_episode_video(env: BaseEnv | BaseParallelEnv | VecEnv,
                          groups: Collection[GroupConfig] = tuple(),
                          policy: AnyPolicyPredictor | PathLike = '',
                          factor: int = 1,
                          seed: int = 0,
                          grouped: bool = False,
                          use_world_bounds: bool = True,
                          of: int = 1,
                          select: int = 0,
                          **kwargs) -> Any:
    """
    Display the video from one episode of an environment.

    :param      env:               The environment
    :param      groups:            The configuration of the groups
    :param      policy:            The default policy
                                   (when not specified in the group config)
    :param      factor:            The real-time factor
    :param      seed:              The seed (only applies if ``of == 0``)
    :param      grouped:           Whether the policy is grouped
    :param      use_world_bounds:  Whether to keep the initial world bounds
    :param      select:            Which episode to select
                                   (ordered by reward, only applies if ``of > 0``)
    :param      of:                The number of runs to choose the episode from
    :param      kwargs:            Keywords arguments passed to the renderer

    """
    exp = make_experiment_with_env(env,
                                   groups=groups,
                                   policy=policy,
                                   record_reward=of > 0,
                                   grouped=grouped,
                                   record_success=False)
    return display_run_video(exp,
                             factor=factor,
                             seed=seed,
                             select=select,
                             of=of,
                             **kwargs)


def record_episode_video(env: BaseEnv | BaseParallelEnv | VecEnv,
                         path: PathLike,
                         groups: Collection[GroupConfig] = tuple(),
                         policy: AnyPolicyPredictor | PathLike = '',
                         factor: int = 1,
                         seed: int = 0,
                         grouped: bool = False,
                         use_world_bounds: bool = True,
                         of: int = 1,
                         select: int = 0,
                         **kwargs):
    """
    Record the video from one episode of an environment.

    :param      env:               The environment
    :param      path:              Where to save the video
    :param      groups:            The configuration of the groups
    :param      policy:            The default policy
                                   (when not specified in the group config)
    :param      factor:            The real-time factor
    :param      seed:              The seed (only applies if ``of == 0``)
    :param      grouped:           Whether the policy is grouped
    :param      use_world_bounds:  Whether to keep the initial world bounds
    :param      select:            Which episode to select
                                   (ordered by reward, only applies if ``of > 0``)
    :param      of:                The number of runs to choose the episode from
    :param      kwargs:            Keywords arguments passed to the renderer

    """
    exp = make_experiment_with_env(env,
                                   groups=groups,
                                   policy=policy,
                                   record_reward=of > 0,
                                   grouped=grouped,
                                   record_success=False)
    record_run_video(exp,
                     path=path,
                     factor=factor,
                     seed=seed,
                     select=select,
                     of=of,
                     **kwargs)



from .base import ActionConfig, ObservationConfig, StateConfig
from .control_action import ControlActionConfig
from .group import GroupConfig, merge_groups_configs
from .modulation_action import ModulationActionConfig
from .observation import DefaultObservationConfig
from .state import DefaultStateConfig
from .discrete_control_action import DiscreteControlActionConfig
from .binary_control_action import BinaryControlActionConfig
from .discrete_control_action_with_comm import DiscreteControlActionWithCommConfig
from .control_action_with_comm import ControlActionWithCommConfig

__all__ = [
    'ControlActionConfig', 'ObservationConfig', 'ModulationActionConfig',
    'GroupConfig', 'ActionConfig', 'merge_groups_configs',
    'DefaultObservationConfig', 'DefaultStateConfig', 'StateConfig',
    'DiscreteControlActionConfig', 'BinaryControlActionConfig',
    'DiscreteControlActionWithCommConfig', 'ControlActionWithCommConfig'
]



from __future__ import annotations

import numpy as np
from stable_baselines3.common.callbacks import BaseCallback
from stable_baselines3.common.utils import safe_mean
from tqdm.auto import tqdm


class ProgressBarWithRewardCallback(BaseCallback):
    """
    Similar to SB3's own
    :py:class:`stable_baselines3.common.callbacks.ProgressBarCallback`, it
    displays a progress bar when training SB3 agent
    using tqdm but includes episodes mean reward and length.
    """
    pbar: tqdm

    def __init__(self, every: int = 1) -> None:
        super().__init__()
        self.every = every

    def _on_training_start(self) -> None:
        # Initialize progress bar
        # Remove timesteps that were done in previous training sessions
        self.pbar = tqdm(total=self.locals["total_timesteps"] -
                         self.model.num_timesteps)
        self.pbar.mininterval = 1
        self._steps = 0
        self._count = 0

    def _on_step(self) -> bool:
        self._steps += self.training_env.num_envs
        # Update progress bar, we do num_envs steps per call to `env.step()`
        # self.pbar.update(self.training_env.num_envs)
        return True

    def _on_rollout_end(self) -> None:
        if self._count % self.every:
            self._count += 1
            return
        if self.model.ep_info_buffer:
            rs = [ep_info["r"] for ep_info in self.model.ep_info_buffer]
            ls = [ep_info["l"] for ep_info in self.model.ep_info_buffer]
            mr = safe_mean(rs)
            sr = 0 if len(rs) == 0 else float(np.std(rs))
            ml = safe_mean(ls)
            sl = 0 if len(ls) == 0 else int(np.std(ls))
            if self.model.ep_success_buffer:
                s = f'({safe_mean(self.model.ep_success_buffer):.2})'
            else:
                s = ''
            self.pbar.set_description(
                f"Reward {mr:.1f} ± {sr:.1f}, Steps {ml:.0f} ± {sl} {s}")
        self.pbar.update(self._steps)
        self._steps = 0
        self._count += 1

    def _on_training_end(self) -> None:
        # Flush and close progress bar
        self.pbar.refresh()
        self.pbar.close()



from __future__ import annotations

import dataclasses as dc
from typing import TYPE_CHECKING, SupportsFloat, cast

import gymnasium as gym
import numpy as np

if TYPE_CHECKING:
    from pettingzoo.utils.env import ParallelEnv

from ...types import Action, AnyPolicyPredictor, Array, Observation
from .env import stack_dict


@dc.dataclass
class Rollout:
    observation: dict[int, Observation]
    action: dict[int, Action]
    reward: dict[int, Array]
    termination: dict[int, bool]
    truncation: dict[int, bool]
    info: dict[int, dict[str, Array]]


def rollout(env: ParallelEnv,
            max_steps: int = 1000,
            policy: AnyPolicyPredictor | None = None,
            deterministic: bool = True,
            seed: int | None = None,
            options: dict = {}) -> Rollout:
    obss: dict[int, list[Observation]] = {i: [] for i in env.possible_agents}
    infos: dict[int, list[dict[str, Array]]] = {
        i: []
        for i in env.possible_agents
    }
    rews: dict[int, list[SupportsFloat]] = {i: [] for i in env.possible_agents}
    acts: dict[int, list[Action]] = {i: [] for i in env.possible_agents}
    obs, info = env.reset(seed=seed, options=options)
    for i, v in obs.items():
        obss[i].append(v)
    for i, v in info.items():
        infos[i].append(v)
    for _ in range(max_steps):
        if policy:
            act = {
                i: policy.predict(o, deterministic=deterministic)[0]
                for i, o in obs.items()
            }
        else:
            act = {i: env.action_space(i).sample() for i in env.agents}
        for i, v in act.items():
            acts[i].append(v)
        obs, rew, term, trunc, info = env.step(act)
        for i, v in obs.items():
            obss[i].append(v)
        for i, v in rew.items():
            rews[i].append(v)
        for i, v in info.items():
            infos[i].append(v)
        if all(term.values()) or any(trunc.values()):
            break
    mobs: dict[int, Observation] = {}
    for i, aobs in obss.items():
        if not aobs:
            continue
        if isinstance(env.observation_space(i), gym.spaces.Dict):
            mobs[i] = stack_dict(cast('list[dict[str, Array]]', aobs))
        else:
            mobs[i] = np.asarray(aobs)
    minfo: dict[int, dict[str, Array]] = {}
    for i, ainfo in infos.items():
        if ainfo:
            minfo[i] = stack_dict(ainfo)
    mrew: dict[int, Array] = {}
    for i, arew in rews.items():
        if arew:
            mrew[i] = np.asarray(arew)
    mact: dict[int, Action] = {}
    for i, aact in acts.items():
        if aact:
            mact[i] = np.asarray(aact)
    return Rollout(observation=mobs,
                   reward=mrew,
                   termination=term,
                   truncation=trunc,
                   info=minfo,
                   action=mact)



from __future__ import annotations

from collections.abc import Iterable, Sequence
from typing import Any, cast

import gymnasium as gym
import numpy as np

from ..indices import Indices, IndicesLike
from ..parallel_env import BaseParallelEnv
from ..types import (Array, Observation, AnyPolicyPredictor, State,
                     accept_info)


def stack_dict(values: dict[Any, Array]) -> Array:
    return np.stack(list(values.values()))


def stack_obs_dict(values: dict[Any, dict[str, Array]],
                   keys: Iterable[str]) -> dict[str, Array]:
    return {
        k: stack_dict({
            i: vs[k]
            for i, vs in values.items()
        })
        for k in keys
    }


def evaluate_policies(
    models: Sequence[tuple[IndicesLike, AnyPolicyPredictor]],
    env: BaseParallelEnv,
    n_eval_episodes: int = 10,
    deterministic: bool = True,
    return_episode_rewards: bool = False,
    warn: bool = True,
) -> tuple[list[float], list[float]] | tuple[list[list[float]], list[int],
                                             list[list[int]]]:
    """
    Mimics StableBaseline3 :py:func:`stable_baselines3.common.evaluation.evaluate_policy`
    on a :py:class:`PettingZoo  Parallel environment <pettingzoo.utils.env.ParallelEnv>`.

    The main differences are:

    - it accepts a list of models to be applied to different sub-groups of agents
    - it returns the rewards divided by groups.

    For example, if a single group is provided

    >>> evaluate_policies(models=[(Indices.all(), model)], env=env)
    ([100.0], [100.0])

    it returns a list of with a single value for the average
    and standard deviation over all episodes of the mean reward over all agents.

    Instead, if we pass two groups, like

    >>> models=[({1, 2}, model1), ({3, 4}, model2), ({5, 6}, model3)]
    >>> evaluate_policies(models=models, env=env)
    ([100.0, 90.0, 110.0], [10.0, 12.0, 7.0])

    it returns two lists of three elements each, one for each group.

    If `return_episode_rewards` is set, it returns three lists:

    - the cumulated rewards for each group and episodes
      (not averaged over the agents!),
      ``[[grp_1_ep_1, grp_1_ep_2, ...], [grp_2_ep_1, grp_2_ep_2, ...]``
    - the length of the episodes
    - the number of agents for each group and episodes.
      ``[[grp_1_ep_1, grp_1_ep_2, ...], [grp_2_ep_1, grp_2_ep_2, ...]``

    For example, with two groups and three episodes, it will be like

    >>> models=[({1, 2}, model1), ({3, 4, 5}, model2)]
    >>> evaluate_policies(models=models, env=env,
                          n_eval_episodes=2,
                          return_episode_rewards=False)
    ([[200.0, 202.0], [300.0, 305.0]], [10, 10], [[2, 2], [3, 3]])


    :param      models:    The models as tuples of indices
                           selecting the agents in the group
                           and the model they will apply

    :param      env:                     The environment

    :param      n_eval_episodes:         The number of episodes

    :param      deterministic:           Whether the policy is applied deterministically

    :param      return_episode_rewards:  Whether to return all episodes (vs averaging them)

    :param      warn:                    Whether to enable warnings

    :returns:   If ``return_episode_rewards`` is set,
        a tuple (list of lists of cumulated episodes rewards',
        list of episodes length,
        list of lists of size of groups),
        else, a tuple (list of average episodes rewards,
        list of std dev of episodes rewards)
    """
    model_indices = [Indices(indices) for indices, _ in models]
    # The cumulated reward of agents in a group
    episode_rewards: list[list[float]] = [[] for _ in models]
    episode_lengths: list[int] = []
    # The number of agents in a group
    episode_numbers: list[list[int]] = [[] for _ in models]
    model_use_info = [accept_info(model.predict) for _, model in models]
    for i in range(n_eval_episodes):
        # The cumulated reward of agents in a group
        observations, infos = env.reset()
        groups = [
            indices.sub_sequence(env.agents) for indices in model_indices
        ]
        # print(env.agents, groups, models)
        current_numbers = [len(group) for group in groups]
        current_reward: list[float] = [0 for group in groups]
        current_length = 0
        group_state: list[State | None] = [None for i in groups]
        done = False
        dones: Array = np.ones((len(env.agents), ), dtype=np.bool_)
        obs_spaces = [
            env.observation_space(group[0]) if group else None
            for group in groups
        ]
        group_obs_keys = [
            space.keys() if isinstance(space, gym.spaces.Dict) else None
            for space in obs_spaces
        ]
        while not done:
            actions = {}
            for i, (group, indices, (_, model), use_info,
                    obs_keys) in enumerate(
                        zip(groups,
                            model_indices,
                            models,
                            model_use_info,
                            group_obs_keys,
                            strict=False)):
                if not group:
                    continue
                group_obs = indices.sub_dict(observations)
                if obs_keys:
                    obs: Observation = stack_obs_dict(
                        cast('dict[Any, dict[str, Array]]', group_obs), obs_keys)
                else:
                    obs = stack_dict(cast('dict[int, Array]', group_obs))
                episode_start = dones[group]
                if use_info:
                    kwargs = {'info': list(indices.sub_dict(infos).values())}
                else:
                    kwargs = {}
                acts, group_state[i] = model.predict(
                    obs,
                    state=group_state[i],
                    episode_start=episode_start,
                    deterministic=deterministic,
                    **kwargs)
                for i, action in zip(group, acts, strict=False):
                    actions[i] = action
            observations, rewards, terminated, truncated, infos = env.step(
                actions)

            for i, indices in enumerate(model_indices):
                current_reward[i] += float(
                    sum(indices.sub_dict(rewards).values()))
            dones = np.bitwise_or(list(terminated.values()),
                                  list(truncated.values()))
            done = bool(np.all(dones))
            current_length += 1

        episode_lengths.append(current_length)
        for i, number in enumerate(current_numbers):
            episode_numbers[i].append(number)
        for i, reward in enumerate(current_reward):
            episode_rewards[i].append(reward)
    if not return_episode_rewards:
        rews = [
            np.asarray(rs) / np.asarray(ns)
            for rs, ns in zip(episode_rewards, episode_numbers, strict=False)
        ]
        mean_reward = [np.mean(rew) for rew in rews]
        std_reward = [np.std(rew) for rew in rews]
        return mean_reward, std_reward
    return episode_rewards, episode_lengths, episode_numbers



from __future__ import annotations

from typing import Any

import numpy as np
from imitation.algorithms import dagger
from imitation.data.rollout import make_sample_until
from tqdm.rich import tqdm

from .rollout import generate_trajectories


# Adds
# - supports for experts callable with info
# - tqdm
class SimpleDAggerTrainer(dagger.SimpleDAggerTrainer):

    def train(self,
              total_timesteps: int,
              *,
              rollout_round_min_episodes: int = 3,
              rollout_round_min_timesteps: int = 500,
              bc_train_kwargs: dict[str, Any] | None = None,
              progress_bar: bool = False) -> None:
        total_timestep_count = 0
        round_num = 0

        if progress_bar:
            bar: tqdm[Any] | None = tqdm(total=total_timesteps)
        else:
            bar = None

        while total_timestep_count < total_timesteps:
            collector = self.create_trajectory_collector()
            round_episode_count = 0
            round_timestep_count = 0

            sample_until = make_sample_until(
                min_timesteps=max(rollout_round_min_timesteps,
                                  self.batch_size),
                min_episodes=rollout_round_min_episodes,
            )

            trajectories = generate_trajectories(
                policy=self.expert_policy,
                venv=collector,
                sample_until=sample_until,
                deterministic_policy=True,
                rng=collector.rng,
            )

            for traj in trajectories:
                self._logger.record_mean(  # type: ignore[no-untyped-call]
                    "dagger/mean_episode_reward",
                    np.sum(traj.rews),
                )
                round_timestep_count += len(traj)
                total_timestep_count += len(traj)

            round_episode_count += len(trajectories)

            self._logger.record(
                "dagger/total_timesteps",
                total_timestep_count)  # type: ignore[no-untyped-call]
            self._logger.record("dagger/round_num",
                                round_num)  # type: ignore[no-untyped-call]
            self._logger.record(
                "dagger/round_episode_count",
                round_episode_count)  # type: ignore[no-untyped-call]
            self._logger.record(
                "dagger/round_timestep_count",
                round_timestep_count)  # type: ignore[no-untyped-call]

            # `logger.dump` is called inside BC.train within the following fn call:
            self.extend_and_update(bc_train_kwargs)
            round_num += 1
            if bar:
                bar.n = min(total_timesteps, total_timestep_count)
                bar.refresh()



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import torch

from benchmarl.experiment.callback import Callback  # type: ignore

if TYPE_CHECKING:
    from tensordict import TensorDictBase  # type: ignore


class ExportPolicyCallback(Callback):
    """
    Export the policy. The best model gets exported as ``"best_policy.onnx"``.
    The others as ``"policy_<iterations>.onnx"``.

    :param export_all: Whether to export all (vs just the best) policies.
    """

    def __init__(self, export_all: bool = False):
        super().__init__()
        self.export_all = export_all

    def on_setup(self):
        self.best_reward = -np.inf

    def on_evaluation_end(self, rollouts: list[TensorDictBase]):
        reward = 0
        for group in self.experiment.group_map:
            reward += np.sum([
                np.sum(rollout['next', group, 'reward'].numpy(), axis=0)
                for rollout in rollouts
            ])
        if reward > self.best_reward:
            for group in self.experiment.group_map:
                self.experiment.export_policy(name="best_policy", group=group)
            self.best_reward = reward
        if self.export_all:
            for group in self.experiment.group_map:
                self.experiment.export_policy(
                    name=f"policy_{self.experiment.n_iters_performed}",
                    group=group)


class AlternateActorCallback(Callback):

    def __init__(self,
                 loss: str,
                 learning_rates: list[tuple[int, float]],
                 exclusive: bool = True,
                 group: str = 'agent'):
        super().__init__()
        self._next_actor_index = 0
        self._next_actor_iter = 0
        self._exclusive = exclusive
        self._learning_rates = learning_rates
        self._loss = loss
        self._group = group

    def on_setup(self) -> None:
        # print(self.experiment.algorithm.get_parameters('agent').keys())
        self.all_params = self.experiment.algorithm.get_parameters(self._group)[
            self._loss]

    # TODO(Jerome): should use `on_train_end` instead
    def on_evaluation_end(self, rollouts: list[TensorDictBase]) -> None:
        if self.experiment.n_iters_performed >= self._next_actor_iter:
            iterations, lr = self._learning_rates[self._next_actor_index]
            self._next_actor_iter = self.experiment.n_iters_performed + iterations
            if self._exclusive:
                i = self._next_actor_index
                params = [{
                    'params': self.all_params[6 * i:6 * (i + 1)],
                    'lr': lr,
                    'weight_decay': 0
                }]
                self.experiment.optimizers[self._group][
                    self._loss] = torch.optim.Adam(params)
            else:
                raise NotImplementedError
            self._next_actor_index = (self._next_actor_index + 1) % len(
                self._learning_rates)



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any

from .base import BaseILAlgorithm

if TYPE_CHECKING:
    from collections.abc import Mapping

    import gymnasium as gym
    import numpy
    from imitation.data.types import Transitions
    from imitation.util.logger import HierarchicalLogger
    from stable_baselines3.common.vec_env import VecEnv

    from . import rollout


def sample_expert_transitions(env: VecEnv,
                              expert: rollout.AnyPolicy,
                              rng: numpy.random.Generator,
                              runs: int = 1000) -> Transitions:
    from imitation.data.rollout import flatten_trajectories, make_sample_until

    from . import rollout

    rollouts = rollout.rollout(
        expert,
        env,
        make_sample_until(min_timesteps=None, min_episodes=runs),
        rng=rng,
    )
    return flatten_trajectories(rollouts)


class BC(BaseILAlgorithm):
    """
    A simplified interface to :py:class:`imitation.algorithms.bc.BC`

    :param env: the environment.
    :param seed: the random seed
    :param policy: an optional policy
    :param policy_kwargs: or the kwargs to create it
    :param logger: an optional logger
    :param runs: how many runs to collect at init
    :param expert: the expert to imitate.
                   If not set, it will default to the policy retrieved
                   from the env attribute "policy" (if set) else to `None`.
    :param bc_kwargs: parameters passed to the
        :py:class:`imitation.algorithms.bc.BC` constructor
    """

    def __init__(self,
                 env: gym.Env[Any, Any] | VecEnv | None = None,
                 seed: int = 0,
                 policy: Any = None,
                 policy_kwargs: Mapping[str, Any] = {'net_arch': [32, 32]},
                 logger: HierarchicalLogger | None = None,
                 runs: int = 0,
                 expert: rollout.AnyPolicy | None = None,
                 bc_kwargs: Mapping[str, Any] = {}):
        import imitation.data.rollout
        from imitation.algorithms import bc

        from . import rollout

        # TODO(Jerome): monkey patching for now
        # In the future, better to subclass BC.
        imitation.data.rollout.generate_trajectories = rollout.generate_trajectories

        super().__init__(env=env,
                         seed=seed,
                         policy=policy,
                         policy_kwargs=policy_kwargs,
                         logger=logger)
        self._expert = expert
        if not expert and self.env:
            try:
                self._expert = self.env.get_attr("policy", [0])[0]
            except (AttributeError, IndexError):
                pass
        self.transitions: Transitions | None = None
        self._bc_kwargs = bc_kwargs
        if runs:
            self.collect_runs(runs)
        self._trainer = bc.BC(
            observation_space=self.observation_space,
            action_space=self.action_space,
            demonstrations=self.transitions,
            rng=self.rng,
            policy=self._policy,
            **self._bc_kwargs,
        )
        self.set_logger(self.logger)

    def set_logger(self, logger: HierarchicalLogger) -> None:
        super().set_logger(logger)
        if self._trainer:
            self._trainer._bc_logger._logger = logger

    def collect_runs(self,
                     runs: int,
                     expert: rollout.AnyPolicy | None = None) -> None:
        """
        Collect training runs from an expert.

        :param runs:   The number of runs
        :param expert: the expert whose trajectories we want to collect.
                       If not set, it will default to the expert
                       configured at init.
        """
        if not expert:
            expert = self._expert
        if not expert:
            warnings.warn("No expert configured", stacklevel=1)
        if expert and self._trainer and self.env and runs > 0:
            self.transitions = sample_expert_transitions(
                self.env, expert, self.rng, runs)
            self._trainer.set_demonstrations(self.transitions)



from __future__ import annotations

import math

from navground import core


class DistributedPadBehavior(core.Behavior, name="Pad"):

    """
    A centralized behavior that applies a (quasi) optimal strategy
    in :py:class:`navground.learning.scenarios.PadScenario`:

    - if both agents have already passed the mid-point of the pad,
      or at least one agent has already passed the pad:
      they advance at full speed
    - if both agents are in the first half of the pad:
      if the fastest way for one agent to exit first from the path
      is to move backwards:
      the agent move forward at full speed, while the
      other keeps advancing at full speed
    - else:
      the agent predicted to exit the pad second, adjust its speed like

      .. code-block::

         speed = min(optimal_speed,
                     distance_to_enter_pad / other_agent_time_to_exit_the_pad)``

      where the computation uses the other agent *current speed* to predict:

      .. code-block::

         other_agent_time_to_exit_the_pad = (
             other_agent_distance_to_exit_pad / other_agent_current_speed)``
    """

    def __init__(self,
                 kinematics: core.Kinematics | None = None,
                 radius: float = 0,
                 pad_width: float = 0.5):
        """
        Constructs a new instance.

        :param      kinematics:  The kinematics
        :param      radius:      The radius
        :param      pad_width:   The pad width
        """
        super().__init__(kinematics=kinematics, radius=radius)
        self._pad_width = pad_width
        self._state = core.GeometricState()

    @property
    def pad_width(self) -> float:
        """
        The pad width

        :returns:   pad width
        """
        return self._pad_width

    @pad_width.setter
    def pad_width(self, value: float) -> None:
        """
        Sets the pad width

        :param      value:  The desired positive value
        """
        self._pad_width = max(0, value)

    def get_environment_state(self) -> core.EnvironmentState:
        return self._state

    def cmd_twist_towards_velocity(self, velocity: core.Vector2,
                                   time_step: float) -> core.Twist2:
        assert len(
            self._state.neighbors) == 1 and self.target.direction is not None
        other = self._state.neighbors[0]
        x = self.target.direction.dot(self.position)
        x_o = -self.target.direction.dot(other.position)
        v_o = abs(self.target.direction.dot(other.velocity))
        b = self.pad_width / 2
        speed = self.optimal_speed
        if x < -b:
            if v_o:
                time_to_enter_o = (-b - x_o) / v_o
            else:
                time_to_enter_o = math.inf
            time_to_enter = (-b - x) / self.optimal_speed
            if x_o > -b or time_to_enter > time_to_enter_o:
                if v_o:
                    time_to_exit_o = (b - x_o) / v_o
                else:
                    time_to_exit_o = math.inf
                if time_to_exit_o > 0:
                    speed = min(self.optimal_speed, (-b - x) / time_to_exit_o)
        elif x < b:
            if abs(x_o) < b:
                # both are inside
                # find minimal time to stop conflict
                if v_o:
                    if x_o > 0:
                        time_to_exit_o = (b - x_o) / v_o
                    else:
                        time_to_exit_o = (b + x_o) / v_o
                else:
                    time_to_exit_o = math.inf
                if x <= 0:
                    time_to_exit = (b + x) / self.optimal_speed
                    if time_to_exit <= time_to_exit_o:
                        speed = -self.optimal_speed
        return core.Twist2((speed, 0), 0, frame=core.Frame.relative)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import numpy as np

from .utils import get_number_of_batches

if TYPE_CHECKING:
    import gymnasium as gym

    from ..types import Action, EpisodeStart, Observation, State


class RandomPredictor:
    """
    This class describes a predictor that returns random actions and
    conforms to :py:class:`navground.learning.types.PolicyPredictor`

    :param action_space:       The action space
    :param observation_space:  An optional observation space
    """
    def __init__(
        self,
        action_space: gym.spaces.Box,
        observation_space: gym.Space[Any],
    ) -> None:
        self.observation_space = observation_space
        self.action_space = action_space

    def get_actions(
        self,
        obs: Observation,
    ) -> Action:
        number = get_number_of_batches(obs, self.observation_space)
        if number == 0:
            return self.action_space.sample()
        return np.stack([self.action_space.sample() for _ in range(number)],
                        axis=0)

    def predict(self,
                observation: Observation,
                state: State | None = None,
                episode_start: EpisodeStart | None = None,
                deterministic: bool = False) -> tuple[Action, State | None]:
        return self.get_actions(observation), None

    def __call__(
        self,
        observation: Observation,
        state: State | None = None,
        episode_start: EpisodeStart | None = None
    ) -> tuple[Action, State | None]:
        return self.predict(observation, state, episode_start, True)



from __future__ import annotations

import copy
import dataclasses as dc
import numbers
from collections import deque, ChainMap
from collections.abc import Collection, Mapping
from typing import Any, Protocol
import warnings

import gymnasium as gym
import numpy as np
from navground import core, sim

from ..config import GroupConfig
from ..types import Array, Reward, TerminationCondition


class ActionProtocol(Protocol):

    def configure(self, behavior: core.Behavior) -> None:
        ...

    def is_configured(self, warn: bool = False) -> bool:
        ...

    @property
    def space(self) -> gym.Space[Any]:
        ...

    def get_cmd_from_action(self, action: Array,
                            behavior: core.Behavior | None,
                            time_step: float) -> core.Twist2:
        ...

    def get_action(self, behavior: core.Behavior, time_step: float) -> Array:
        ...


class ObservationProtocol(Protocol):

    def configure(self, behavior: core.Behavior | None,
                  sensing_space: gym.spaces.Dict) -> None:
        ...

    def is_configured(self, warn: bool = False) -> bool:
        ...

    def get_observation(
            self, behavior: core.Behavior | None,
            buffers: Mapping[str, core.Buffer]) -> dict[str, Array] | Array:
        ...

    @property
    def space(self) -> gym.Space[Any]:
        ...

    @property
    def history(self) -> int:
        ...


def make_simple_space(desc: core.BufferDescription) -> gym.Space[Any]:
    is_int = issubclass(desc.type.type, numbers.Integral)
    if is_int and desc.low == 0 and desc.high == 1:
        # CHANGED it does not respect the sensor dtype but instead uses int8
        # return gym.spaces.MultiBinary(desc.shape)
        pass
    if desc.categorical and is_int and len(desc.shape) == 1:
        return gym.spaces.MultiDiscrete(nvec=[int(desc.high - desc.low + 1)] *
                                        desc.shape[0],
                                        dtype=desc.type.type,
                                        start=[int(desc.low)] * desc.shape[0])
    return gym.spaces.Box(desc.low,
                          desc.high,
                          desc.shape,
                          dtype=desc.type.type)


def make_composed_space(
        value: Mapping[str, core.BufferDescription]) -> gym.spaces.Dict:
    return gym.spaces.Dict({
        k: make_simple_space(desc)
        for k, desc in value.items()
    })


def get_relative_target_position(behavior: core.Behavior) -> core.Vector2:
    if behavior.target.position is None:
        return np.zeros(2, dtype=core.FloatType)
    return core.to_relative(behavior.target.position - behavior.position,
                            behavior.pose)


def get_space_for_sensor(sensor: sim.Sensor) -> gym.spaces.Dict:
    return make_composed_space(sensor.description)


def get_space_for_sensors(sensors: list[sim.Sensor]) -> gym.spaces.Dict:
    return make_composed_space(ChainMap(*[sensor.description for sensor in sensors]))


def get_space_for_state(state: core.SensingState) -> gym.spaces.Dict:
    return make_composed_space({
        k: v.description
        for k, v in state.buffers.items()
    })


def get_state(behavior: core.Behavior | None,
              state: core.SensingState | None) -> core.SensingState | None:
    if state is None and behavior and isinstance(behavior.environment_state,
                                                 core.SensingState):
        return behavior.environment_state
    return state


class GymAgent:

    def __init__(self,
                 action: ActionProtocol,
                 observation: ObservationProtocol,
                 behavior: core.Behavior | None = None,
                 state: core.SensingState | None = None):
        state = get_state(behavior, state)
        if state:
            sensing_space = get_space_for_state(state)
        else:
            sensing_space = gym.spaces.Dict({})
        self.observation_config = copy.copy(observation)
        self.action_config = copy.copy(action)
        if behavior:
            self.action_config.configure(behavior)
        self.observation_config.configure(behavior, sensing_space)
        self.init(behavior, state)

    def is_configured(self, warn: bool = False) -> bool:
        return (self.action_config.is_configured(warn)
                and self.observation_config.is_configured(warn))

    @property
    def action_space(self) -> gym.Space[Any]:
        return self.action_config.space

    @property
    def observation_space(self) -> gym.Space[Any]:
        return self.observation_config.space

    def init(self, behavior: core.Behavior | None,
             state: core.SensingState | None) -> None:
        self._behavior = behavior
        state = get_state(behavior, state)
        if state:
            self._buffers = dict(state.buffers)
        else:
            self._buffers = {}
        history = self.observation_config.history
        if history > 1:
            self._stack: deque[dict[str, Array]
                               | Array] | None = deque(maxlen=history)
        else:
            self._stack = None

    def get_cmd_from_action(self, action: Array,
                            time_step: float) -> core.Twist2:
        return self.action_config.get_cmd_from_action(action, self._behavior,
                                                      time_step)

    def get_action(self, time_step: float) -> Array:
        assert self._behavior is not None
        return self.action_config.get_action(self._behavior, time_step)

    def update_observation(self) -> dict[str, Array] | Array:
        fs = self.observation_config.get_observation(self._behavior,
                                                     self._buffers)
        if self._stack is None:
            return fs
        self._stack.append(fs)
        while len(self._stack) < self.observation_config.history:
            self._stack.append(fs)
        return np.asarray(self._stack)

    # def reset(self) -> None:
    #     if self._stack:
    #         self._stack.clear()

    def __getstate__(self) -> tuple[Any, ...]:
        return (self.action_config, self.observation_config)

    def __setstate__(self, state: tuple[Any, ...]) -> None:
        (self.action_config, self.observation_config) = state
        self.init(None, None)


@dc.dataclass
class Agent:
    gym: GymAgent | None = None
    reward: Reward | None = None
    navground: sim.Agent | None = None
    state: core.SensingState | None = None
    sensors: list[sim.Sensor] = dc.field(default_factory=list)
    policy: Any = None
    terminate_on_success: bool = True
    terminate_on_failure: bool = True
    success_condition: TerminationCondition | None = None
    failure_condition: TerminationCondition | None = None

    def get_sensors(self) -> list[sim.Sensor]:
        if self.sensors:
            return self.sensors
        if self.navground:
            return [se for se in self.navground.state_estimations if isinstance(se, sim.Sensor)]
        return []

    def update_state(self, world: sim.World) -> None:
        if self.sensors and self.state and self.navground:
            for sensor in self.sensors:
                sensor.update(self.navground, world, self.state)

    def is_configured(self, warn: bool = False) -> bool:
        if self.gym:
            return self.gym.is_configured(warn)
        if warn:
            warnings.warn("No Gymnasium configuration", stacklevel=1)
        return False


def create_agents_in_group(
        world: sim.World,
        group: GroupConfig,
        max_number_of_agents: int | None = None) -> dict[int, Agent]:
    agents = {}
    sensors = group.get_sensors()
    world_agents = group.indices.sub_sequence(world.agents)
    if world_agents:
        representative_behavior = world_agents[0].behavior
    else:
        representative_behavior = None
    if max_number_of_agents is not None and max_number_of_agents > len(
            world.agents):
        number_of_agents = max_number_of_agents
    else:
        number_of_agents = len(world.agents)
    indices = group.indices.as_set(number_of_agents)
    for i in indices:
        if sensors:
            state = core.SensingState()
            for sensor in sensors:
                sensor.prepare_state(state)
        else:
            state = None
        if i >= 0 and i < len(world.agents):
            ng_agent = world.agents[i]
            behavior: core.Behavior | None = ng_agent.behavior
        else:
            ng_agent = None
            behavior = representative_behavior
        if group.observation and group.action:
            gym_agent = GymAgent(observation=group.observation,
                                 action=group.action,
                                 behavior=behavior,
                                 state=state)

        else:
            gym_agent = None
        if ng_agent and group.color:
            ng_agent.color = group.color
        if ng_agent and group.tag:
            ng_agent.add_tag(group.tag)

        if group.terminate_on_success is not None:
            terminate_on_success = group.terminate_on_success
        else:
            terminate_on_success = True

        if group.terminate_on_failure is not None:
            terminate_on_failure = group.terminate_on_failure
        else:
            terminate_on_failure = True

        agents[i] = Agent(state=state,
                          gym=gym_agent,
                          sensors=sensors,
                          reward=group.reward,
                          navground=ng_agent,
                          terminate_on_success=terminate_on_success,
                          terminate_on_failure=terminate_on_failure,
                          success_condition=group.success_condition,
                          failure_condition=group.failure_condition)
        # TODO(Jerome):
        # Set policy
    return agents


def create_agents_in_groups(
        world: sim.World,
        groups: Collection[GroupConfig],
        max_number_of_agents: int | None = None) -> dict[int, Agent]:
    agents = {}
    for group in groups:
        agents.update(
            create_agents_in_group(world, group, max_number_of_agents))
    return agents



"""
Modified from https://github.com/HerveMignot/switch_magic
MIT Licence, Copyright (c) 2018 Hervé Mignot
"""

from __future__ import annotations
import gc

from IPython.core.magic import register_cell_magic
from IPython import get_ipython


@register_cell_magic
def skip_if(line: str, cell: str) -> None:
    """
    Executes the cell if the condition in ``line`` evaluate to False.

    See `switch_magic <https://github.com/HerveMignot/switch_magic>`_.

    :param line:  The condition
    :param cell:  The cell
    """
    if eval(line):
        return
    get_ipython().run_cell(cell)


@register_cell_magic
def run_if(line: str, cell: str) -> None:
    """
    Executes the cell if the condition in ``line`` evaluate to True.

    See `switch_magic <https://github.com/HerveMignot/switch_magic>`_.

    :param line:  The condition
    :param cell:  The cell
    """
    if not eval(line):
        return
    get_ipython().run_cell(cell)


@register_cell_magic
def run_and_time_if(line: str, cell: str) -> None:
    """
    Executes the cell if the condition in ``line`` evaluate to True
    and time it.

    See `switch_magic <https://github.com/HerveMignot/switch_magic>`_.

    :param line:  The condition
    :param cell:  The cell
    """
    if not eval(line):
        return
    get_ipython().magics_manager.magics['cell']['time'](cell=cell)


def clean_tqdm_rich() -> None:
    """
    Force closing any tqdm rich object
    """
    for obj in gc.get_objects():
        if 'tqdm_rich' in type(obj).__name__:
            try:
                obj.close()
            except Exception:
                pass



from __future__ import annotations

import functools
from collections.abc import Iterable, Mapping, Sequence
from enum import IntEnum
from operator import itemgetter
from typing import Any, Literal, TypeAlias, TypeVar, cast

T = TypeVar("T")


class Indices:
    """
    A union type that can be used to get a part of a sequence or
    of a dictionary with integer keys.

    - :py:attr:`Indices.all` -> select all items
    - :py:class:`set` -> select the items with indices in the set
    - :py:class:`slice` -> select the items covered by the slide.

    For example:

    >>> Indices.all().sub_sequence([1, 2, 3])
    [1, 2, 3]
    >>> Indices.all().sub_dict({1: 2, 3: 4, 5: 6})
    {1: 2, 3: 4, 5: 6}
    >>> Indices({0, 2}).sub_sequence((1, 2, 3))
    (1, 3)
    >>> Indices({0, 2}).sub_dict({1: 2, 3: 4, 5: 6})
    {}
    >>> Indices(slice(-2, None)).sub_sequence((1, 2, 3, 4, 5))
    (4, 5)
    >>> Indices(slice(1, 3)).sub_dict({1: 2, 3: 4, 5: 6})
    {1: 2}

    :param      value:  The value
        - ``"ALL"``: initializes :py:attr:`Indices.all`
        - :py:class:`set`: initializes the indices from the set
        - :py:class:`slice`: initializes the indices from the slide.
        - :py:class:`Indices`: copy the value
    """

    __slots__ = '_type', '_value'

    class Type(IntEnum):
        slice = 0
        set = 1
        all = 2

    @classmethod
    def all(cls) -> Indices:
        """
        Indices that covers all the items.
        """
        return Indices()

    def __init__(self, value: IndicesLike = "ALL"):
        self._value: slice | set[int] | Literal['ALL']
        self._type: Indices.Type
        if isinstance(value, Indices):
            self._value = value._value
            self._type = value._type
        elif isinstance(value, slice):
            self._type = Indices.Type.slice
            self._value = value
        elif isinstance(value, (list, tuple, set)):
            self._type = Indices.Type.set
            self._value = set(value)
        else:
            self._type = Indices.Type.all
            self._value = "ALL"

    def __repr__(self) -> str:
        if self._type == Indices.Type.all:
            return "Indices.all()"
        return f"Indices({self._value!r})"

    def __hash__(self) -> int:
        return hash(self._value)

    def sub_sequence(self, xs: Sequence[T]) -> Sequence[T]:
        """
        Extracts a sub-sequence.

        :param      xs:  The original sequence

        :returns:   the items of xs covered by the indices
        """
        if self._type == Indices.Type.all:
            return xs
        if self._type == Indices.Type.set:
            ys = cast("set[int]", self._value) & set(range(0, len(xs)))
            if ys:
                z = itemgetter(*ys)(xs)
                if len(ys) > 1:
                    return cast("Sequence[T]", z)
                return type(xs)([z])  # type: ignore[call-arg]

            return type(xs)()
        return xs[cast("slice", self._value)]

    def sub_dict(self, xs: dict[int, T]) -> dict[int, T]:
        """
        Extracts a sub-set.

        :param      xs:  The original sequence

        :returns:   the items of xs covered by the indices
        """
        if self._type == Indices.Type.all:
            return xs
        return {k: v for k, v in xs.items() if k in self.as_set(max(xs) + 1)}

    def as_set(self, length: int) -> set[int]:
        """
        Covert to a set of the indices.
        Uses length to evaluate slices using :py:meth:`slice.indices`.

        :param      length:  The maximal length of the sequence

        :returns:   the set of indices
        """
        if self._type == Indices.Type.set:
            return cast("set[int]", self._value)
        if self._type == Indices.Type.all:
            return set(range(0, length))
        return set(range(*cast("slice", self._value).indices(length)))

    @property
    def lowest(self) -> int | None:
        """
        The lowest index

        :returns:   The lowest index if the available, else ``None``.
        """
        if self._type == Indices.Type.all:
            return 0
        if self._type == Indices.Type.set:
            if self._value:
                return min(cast("set[int]", self._value))
            return None
        a, b, _ = cast("slice", self._value).indices(1)
        if a < b:
            return a
        return None

    @property
    def asdict(self) -> dict[str, Any]:
        """
        A JSON-able representation.

        :returns:   A JSON-able dict
        """
        rs: dict[str, Any] = {}
        rs['type'] = self._type.name
        if self._type == Indices.Type.set:
            rs['values'] = list(cast("set[int]", self._value))
        elif self._type == Indices.Type.slice:
            rs['start'] = cast("slice", self._value).start
            rs['stop'] = cast("slice", self._value).stop
            rs['step'] = cast("slice", self._value).step
        return rs

    @classmethod
    def from_dict(cls, rs: Mapping[str, Any]) -> Indices:
        """
        Read the indices from the representation :py:attr:`asdict`.

        :param      rs:   The dictionart

        :returns:   The indices
        :raises ValueError: if no conversion is possible
        """
        if rs['type'] == 'all':
            return cls.all()
        if rs['type'] == 'set':
            return cls(set(rs['values']))
        if rs['type'] == 'slice':
            return cls(slice(rs['start'], rs['stop'], rs['step']))
        raise ValueError(f"Could not read indices from {rs}")

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Indices):
            return False
        return self._value == other._value

    def intersect(self, other: Indices, length: int) -> Indices:
        """
        Intersects with another set of indices.

        :param      other:   The other set
        :param      length:  The maximal length of the sequences, used to evaluate slices.

        :returns:   The set of indices that are common to both sets.
        """
        if self == other:
            return self
        if self._type == Indices.Type.all:
            return other
        return Indices(other.as_set(length) & self.as_set(length))

    def __bool__(self) -> bool:
        """
        :returns:   True if not empty.
        """
        if self._type == Indices.Type.all:
            return True
        if self._type == Indices.Type.set:
            return bool(cast("set[int]", self._value))
        s = cast("slice", self._value)
        start = s.start or 0
        step = s.step or 1
        stop = s.stop or (start + step)
        return (stop - start) * step > 0


def join_indices(values: Iterable[Indices], length: int) -> Indices:
    """
    Returns the union indices

    :param      values:  The values
    :param      length:  The maximal length of the sequences, used to evaluate slices.


    :returns:   Indices containing all the indices in values.
    """
    vs: list[set[int]] = [indices.as_set(length) for indices in values]
    if vs:
        iss: set[int] = functools.reduce(lambda x, y: x | y, vs)
    else:
        iss = set()
    return Indices(iss)


IndicesLike: TypeAlias = Indices | slice | list[int] | tuple[int] | set[int] | Literal[
    'ALL']



def test_import():
    import navground.learning.utils.benchmarl  # noqa


def test_create_env():
    from navground.learning.examples.cross import get_env
    from navground.learning.utils.benchmarl import make_env

    penv = get_env(multi_agent=True)
    env = make_env(env=penv, seed=0, categorical_actions=False)
    env.set_seed(0)
    env.reset()



from __future__ import annotations

from collections.abc import Sequence

import torch as th

from ..types import PyTorchPolicy


class OnnxablePolicy(th.nn.Module):

    def __init__(self, policy: PyTorchPolicy):
        super().__init__()
        self.policy = policy

    def forward(self, observation: th.Tensor) -> th.Tensor:
        return self.policy(observation, deterministic=True)


class OnnxablePolicyWithMultiInput(th.nn.Module):

    def __init__(self, policy: PyTorchPolicy, keys: Sequence[str]):
        super().__init__()
        self.policy = policy
        self.keys = keys

    def forward(self, *observations: th.Tensor) -> th.Tensor:
        obs = dict(zip(self.keys, list(observations), strict=True))
        return self.policy(obs, deterministic=True)



from __future__ import annotations

from typing import Any

import gymnasium as gym
import numpy as np

from ..types import Action, EpisodeStart, Info, Observation, State


class InfoPolicy:
    """
    A predictor that extracts navground actions
    from the info dictionary and conforms to
    :py:class:`navground.learning.types.PolicyPredictorWithInfo`

    :param action_space:       The action space
    :param key:                The key of the action in the ``info`` dictionary
    :param observation_space:  An optional observation space
    """

    def __init__(self,
                 action_space: gym.Space[Any],
                 key: str,
                 observation_space: gym.Space[Any] = gym.spaces.Dict()) -> None:
        self.observation_space = observation_space
        self.action_space = action_space
        self.key = key
        self._default = self.action_space.sample() * 0

    def get_actions(self, info: Info) -> Action:
        if isinstance(info, list):
            return np.asarray([i.get(self.key, self._default) for i in info])
        return info.get(self.key, self._default)

    def predict(self,
                observation: Observation,
                state: State | None = None,
                episode_start: EpisodeStart | None = None,
                deterministic: bool = False,
                info: Info | None = None) -> tuple[Action, State | None]:
        assert info is not None
        return self.get_actions(info), None

    def __call__(self,
                 observation: Observation,
                 state: State | None = None,
                 episode_start: EpisodeStart | None = None,
                 info: Info | None = None) -> tuple[Action, State | None]:
        return self.predict(observation, state, episode_start, True, info)



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import torch as th
from stable_baselines3.common.policies import BasePolicy

if TYPE_CHECKING:
    from stable_baselines3.common.type_aliases import PyTorchObs


class NullPolicy(BasePolicy):
    """
    This class describes a dummy policy that always returns zeros.
    """
    def _predict(
        self,
        observation: PyTorchObs,
        deterministic: bool = False,
    ) -> th.Tensor:
        if isinstance(observation, dict):
            number = next(iter(observation.values())).shape[0]
        else:
            number = observation.shape[0]
        action = np.zeros(
            shape=(number, *self.action_space.shape),  # type: ignore[misc]
            dtype=self.action_space.dtype)
        return th.as_tensor(action)

    def forward(self, x: PyTorchObs, deterministic: bool = False) -> th.Tensor:
        return self._predict(x, deterministic=deterministic)



from __future__ import annotations

import pathlib
from typing import cast

try:
    from typing import Self
except ImportError:
    try:
        from typing_extensions import Self
    except ImportError:
        ...

from navground import core

from ...config import ControlActionConfig, DefaultObservationConfig
from ...internal.group import GymAgent
from ...types import (Action, AnyPolicyPredictor, Observation,
                      ObservationTransform, PathLike)


class BasePolicyMixin:
    """
    A navigation behavior that evaluates a ML policy

    *Registered properties*:

    - :py:attr:`policy_path` (str)
    - :py:attr:`flat` (bool)
    - :py:attr:`history` (int)
    - :py:attr:`fix_orientation` (bool)
    - :py:attr:`include_target_direction` (bool)
    - :py:attr:`include_target_direction_validity` (bool)
    - :py:attr:`include_target_distance` (bool)
    - :py:attr:`include_target_distance_validity` (bool)
    - :py:attr:`include_target_speed` (bool)
    - :py:attr:`include_target_angular_speed` (bool)
    - :py:attr:`include_velocity` (bool)
    - :py:attr:`include_angular_speed` (bool)
    - :py:attr:`include_radius` (bool)
    - :py:attr:`use_wheels` (bool)
    - :py:attr:`use_acceleration_action` (bool)
    - :py:attr:`max_acceleration` (float)
    - :py:attr:`max_angular_acceleration` (float)
    - :py:attr:`deterministic` (bool)

    *State*: :py:class:`core.SensingState`

    :param kinematics: The agent kinematics
    :param radius: The agent radius
    :param policy: The policy
    :param policy_path: The policy path
    :param action_config: Configures which actions the policy generates
    :param observation_config: Configures which observations the policy consumes
    :param deterministic: Whether the policy evaluation is deterministic
    :param pre: Optional input (observations) transformation
    """

    _policies: dict[pathlib.Path, AnyPolicyPredictor] = {}
    _cache_is_enabled: bool = False

    def __init__(self,
                 policy: AnyPolicyPredictor | None = None,
                 policy_path: PathLike = '',
                 action_config: ControlActionConfig = ControlActionConfig(),
                 observation_config:
                 DefaultObservationConfig = DefaultObservationConfig(),
                 deterministic: bool = False,
                 pre: ObservationTransform | None = None):
        self._state = core.SensingState()
        self._policy = policy
        self._policy_path: pathlib.Path | None = None
        self._gym_agent: GymAgent | None = None
        self._action_config = action_config
        self._observation_config = observation_config
        self._deterministic = deterministic
        self._pre = pre

    @classmethod
    def enable_cache(cls, value: bool) -> None:
        cls._cache_is_enabled = value

    @classmethod
    def cache_is_enabled(cls) -> bool:
        return cls._cache_is_enabled

    @classmethod
    def reset_cache(cls) -> None:
        cls._policies.clear()

    def init_policy(self) -> None:
        """
        Loads the policy from :py:attr:`policy_path` if not already done.
        """
        if not self._policy and self._policy_path:
            self._policy = self.load_policy(self._policy_path)

    @classmethod
    def load_policy(cls, path: str | pathlib.Path) -> AnyPolicyPredictor:
        if isinstance(path, str):
            path = pathlib.Path(path)
        if cls.cache_is_enabled():
            path = path.absolute()
            if path not in cls._policies:
                cls._policies[path] = cls._load_policy(path)
            return cls._policies[path]
        return cls._load_policy(path)

    @classmethod
    def _load_policy(cls, path: pathlib.Path) -> AnyPolicyPredictor:
        if path.suffix == '.onnx':
            return cls._load_onnx_policy(path)
        return cls._load_torch_policy(path)

    @classmethod
    def _load_onnx_policy(cls, path: pathlib.Path) -> AnyPolicyPredictor:
        from ...onnx.policy import OnnxPolicy

        return OnnxPolicy(path)

    @classmethod
    def _load_torch_policy(cls, path: pathlib.Path) -> AnyPolicyPredictor:
        import torch
        from stable_baselines3.common.policies import BasePolicy
        try:
            model = torch.load(path, weights_only=False)
            if isinstance(model, BasePolicy):
                return model
            raise RuntimeError("Loaded model is not a policy")
        except Exception as e:
            raise RuntimeError(f"Could not load policy from {path}") from e

    @property
    @core.register("", "policy path")
    def policy_path(self) -> str:
        """
        The file from which to load the model
        """
        if not self._policy_path:
            return ''
        return str(
            self._policy_path.relative_to(pathlib.Path.cwd(), walk_up=True))
        # return str(self._policy_path)

    @policy_path.setter
    def policy_path(self, value: PathLike) -> None:
        self.set_policy_path(value, load_policy=False)

    def set_policy_path(self,
                        value: PathLike,
                        load_policy: bool = False) -> None:
        value = pathlib.Path(value).absolute()
        try:
            if self._policy_path and self._policy_path.samefile(value):
                return
        except FileNotFoundError:
            return
        self._policy_path = value
        self._policy = None
        if load_policy:
            self.init_policy()

    @property
    @core.register(
        False, "Whether to include the target distance in the observations")
    def include_target_distance(self) -> bool:
        """
        See :py:attr:`.DefaultObservationConfig.include_target_distance`
        """
        return self._observation_config.include_target_distance

    @include_target_distance.setter
    def include_target_distance(self, value: bool) -> None:
        self._observation_config.include_target_distance = value

    @property
    @core.register(
        False,
        "Whether to include the target distance validity in the observations")
    def include_target_distance_validity(self) -> bool:
        """
        See :py:attr:`.DefaultObservationConfig.include_target_distance_validity`
        """
        return self._observation_config.include_target_distance_validity

    @include_target_distance_validity.setter
    def include_target_distance_validity(self, value: bool) -> None:
        self._observation_config.include_target_distance_validity = value

    @property
    @core.register(
        True, "Whether to include the target direction in the observations")
    def include_target_direction(self) -> bool:
        """
        See :py:attr:`.DefaultObservationConfig.include_target_direction`
        """
        return self._observation_config.include_target_direction

    @include_target_direction.setter
    def include_target_direction(self, value: bool) -> None:
        self._observation_config.include_target_direction = value

    @property
    @core.register(
        False,
        "Whether to include the target direction validity in the observations")
    def include_target_direction_validity(self) -> bool:
        """
        See :py:attr:`.DefaultObservationConfig.include_target_direction_validity`
        """
        return self._observation_config.include_target_direction_validity

    @include_target_direction_validity.setter
    def include_target_direction_validity(self, value: bool) -> None:
        self._observation_config.include_target_direction_validity = value

    @property
    @core.register(False,
                   "Whether to include the target speed in the observations")
    def include_target_speed(self) -> bool:
        """
        See :py:attr:`.DefaultObservationConfig.include_target_speed`
        """
        return self._observation_config.include_target_speed

    @include_target_speed.setter
    def include_target_speed(self, value: bool) -> None:
        self._observation_config.include_target_speed = value

    @property
    @core.register(False,
                   "Whether to include the target speed in the observations")
    def include_target_angular_speed(self) -> bool:
        """
        See :py:attr:`.DefaultObservationConfig.include_target_angular_speed`
        """
        return self._observation_config.include_target_angular_speed

    @include_target_angular_speed.setter
    def include_target_angular_speed(self, value: bool) -> None:
        self._observation_config.include_target_angular_speed = value

    @property
    @core.register(
        False, "Whether to include the current velocity in the observations")
    def include_velocity(self) -> bool:
        """
        See :py:attr:`.DefaultObservationConfig.include_velocity`
        """
        return self._observation_config.include_velocity

    @include_velocity.setter
    def include_velocity(self, value: bool) -> None:
        self._observation_config.include_velocity = value

    @property
    @core.register(
        False,
        "Whether to include the current angular speed in the observations")
    def include_angular_speed(self) -> bool:
        """
        See :py:attr:`.DefaultObservationConfig.include_angular_speed`
        """
        return self._observation_config.include_angular_speed

    @include_angular_speed.setter
    def include_angular_speed(self, value: bool) -> None:
        self._observation_config.include_angular_speed = value

    @property
    @core.register(False,
                   "Whether to include the own radius in the observations")
    def include_radius(self) -> bool:
        """
        See :py:attr:`.DefaultObservationConfig.include_radius`
        """
        return self._observation_config.include_radius

    @include_radius.setter
    def include_radius(self, value: bool) -> None:
        self._observation_config.include_radius = value

    @property
    @core.register(False, "Whether to flatten the observations")
    def flat(self) -> bool:
        """
        See :py:attr:`.DefaultObservationConfig.flat`
        """
        return self._observation_config.flat

    @flat.setter
    def flat(self, value: bool) -> None:
        self._observation_config.flat = value

    @property
    @core.register(1, "Length of the queue")
    def history(self) -> int:
        """
        See :py:attr:`.DefaultObservationConfig.history`
        """
        return self._observation_config.history

    @history.setter
    def history(self, value: int) -> None:
        self._observation_config.history = value

    @property
    @core.register(False, "Whether to keep orientation fixed")
    def fix_orientation(self) -> bool:
        """
        See :py:attr:`.ControlActionConfig.fix_orientation`
        """
        return self._action_config.fix_orientation

    @fix_orientation.setter
    def fix_orientation(self, value: bool) -> None:
        self._action_config.fix_orientation = value

    @property
    @core.register(
        False,
        "Whether action and observation uses wheel speeds or accelerations")
    def use_wheels(self) -> bool:
        """
        See :py:attr:`.ControlActionConfig.use_wheels`
        """
        return self._action_config.use_wheels

    @use_wheels.setter
    def use_wheels(self, value: bool) -> None:
        self._action_config.use_wheels = value

    @property
    @core.register(False, "Whether actions are accelerations.")
    def use_acceleration_action(self) -> bool:
        """
        See :py:attr:`.ControlActionConfig.use_acceleration_action`
        """
        return self._action_config.use_acceleration_action

    @use_acceleration_action.setter
    def use_acceleration_action(self, value: bool) -> None:
        self._action_config.use_acceleration_action = value

    @property
    @core.register(10.0, "The upper bound of the acceleration.")
    def max_acceleration(self) -> float:
        """
        See :py:attr:`.ControlActionConfig.max_acceleration`
        """
        return self._action_config.max_acceleration

    @max_acceleration.setter
    def max_acceleration(self, value: float) -> None:
        self._action_config.max_acceleration = value

    @property
    @core.register(100.0, "The upper bound of the angular acceleration.")
    def max_angular_acceleration(self) -> float:
        """
        See :py:attr:`.ControlActionConfig.max_angular_acceleration`
        """
        return self._action_config.max_angular_acceleration

    @max_angular_acceleration.setter
    def max_angular_acceleration(self, value: float) -> None:
        self._action_config.max_angular_acceleration = value

    @property
    @core.register(False, "Whether or not to output deterministic actions")
    def deterministic(self) -> bool:
        """
        Whether or not to output deterministic actions
        """
        return self._deterministic

    @deterministic.setter
    def deterministic(self, value: bool) -> None:
        self._deterministic = value

    def get_environment_state(self) -> core.SensingState:
        return self._state

    def prepare(self) -> None:
        if not self._gym_agent:
            self._gym_agent = GymAgent(observation=self._observation_config,
                                       action=self._action_config,
                                       behavior=cast('core.Behavior', self))

    def update_observation(self) -> Observation:
        if not self._gym_agent:
            raise RuntimeError("Not configured")
        return self._gym_agent.update_observation()

    def get_cmd_from_action(self, action: Action,
                            time_step: float) -> core.Twist2:
        if not self._gym_agent:
            raise RuntimeError("Not configured")
        return cast('core.Behavior', self).feasible_twist(
            self._gym_agent.get_cmd_from_action(action, time_step))

    @classmethod
    def clone_behavior(cls,
                       behavior: core.Behavior,
                       policy: AnyPolicyPredictor | PathLike | None,
                       action_config: ControlActionConfig,
                       observation_config: DefaultObservationConfig,
                       deterministic: bool = False,
                       pre: ObservationTransform | None = None) -> Self:
        """
        Configure a new policy behavior from a behavior.

        :param      behavior:         The behavior to replicate
        :param      policy:           The policy
        :param      config:           The configuration
        :param      deterministic:    Whether or not to output deterministic actions
        :param pre: Optional input (observations) transformation

        :returns:   The configured policy behavior
        """
        try:
            policy_path: PathLike = pathlib.Path(
                policy)  # type: ignore[arg-type]
            policy = None
        except TypeError:
            policy_path = ''
        pb = cls(policy=cast("AnyPolicyPredictor | None", policy),
                 action_config=action_config,
                 observation_config=observation_config,
                 deterministic=deterministic,
                 pre=pre)
        if policy_path:
            pb.set_policy_path(policy_path, load_policy=True)
        cast('core.Behavior', pb).set_state_from(behavior)
        return pb

    def clone(self) -> Self:
        """
        Creates a new policy behavior with same properties but a separate state.

        :returns:   Copy of this object.
        """
        return type(self).clone_behavior(
            behavior=cast('core.Behavior', self),
            policy=self._policy,
            action_config=self._action_config,
            observation_config=self._observation_config,
            deterministic=self._deterministic)



from __future__ import annotations

import numpy as np

from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:
    from stable_baselines3.common.vec_env import VecEnv
    import gymnasium as gym


def maybe_make_venv(
    env: gym.Env[Any, Any] | VecEnv,
    parallel: bool = False,
    num_envs: int = 1,
    rng: np.random.Generator = np.random.default_rng(0)
) -> VecEnv:
    from stable_baselines3.common.vec_env import VecEnv
    if isinstance(env, VecEnv):
        return env
    return make_vec_from_env(env, parallel, num_envs, rng)


def make_vec_from_env(
    env: gym.Env[Any, Any],
    parallel: bool = False,
    num_envs: int = 1,
    rng: np.random.Generator = np.random.default_rng(0)
) -> VecEnv:
    """
    Creates an imitation-compatible vectorized enviroment.
    Just a tiny wrapped on :py:func:`imitation.util.util.make_vec_env`
    that reads ``env_id`` and  ``env_make_kwargs`` from the env.

    :param      env:       The environment
    :param      parallel:  Whether to run the venv in parallel
    :param      number:    The number of environments
    :param      rng:       The random number generator

    :returns:   The vectorized environment.
    """
    from imitation.util.util import make_vec_env
    from imitation.data.wrappers import RolloutInfoWrapper

    if not env.unwrapped.spec:
        raise ValueError("env has an empty spec")
    env_id = env.unwrapped.spec.id
    env_kwargs = env.unwrapped.spec.kwargs
    return make_vec_env(env_id,
                        rng=rng,
                        post_wrappers=[lambda env, _: RolloutInfoWrapper(env)],
                        env_make_kwargs=env_kwargs,
                        parallel=parallel,
                        n_envs=num_envs)



from __future__ import annotations

# import copy
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from benchmarl.models.common import ModelConfig  # type: ignore[import-not-found]
    from benchmarl.algorithms.common import AlgorithmConfig  # type: ignore[import-not-found]
    from benchmarl.experiment import Callback, ExperimentConfig  # type: ignore[import-not-found]
    from ...types import PathLike
    from ...onnx.policy import OnnxPolicy
    import pandas as pd
    import gymnasium as gym

import math
from pathlib import Path

from benchmarl.experiment import Experiment  # type: ignore[import-not-found]

from ...config import GroupConfig
from ...indices import Indices
from ...parallel_env import MultiAgentNavgroundEnv
from .evaluate import evaluate_policy
from .navground_task import NavgroundTaskClass
from .policy import SingleAgentPolicy


class NavgroundExperiment(Experiment):
    """
    A :py:class:`benchmarl.experiment.Experiment` created from
    a :py:class:`navground.learning.parallel_env.MultiAgentNavgroundEnv`.

    :param env:                 The training environment

    :param task:                The task

    :param algorithm_config:    The algorithm configuration

    :param model_config:        The model configuration

    :param seed:                The seed

    :param config:              The experiment configuration

    :param critic_model_config: The critic model configuration

    :param callbacks:           The callbacks

    :param eval_env:            The evaluation environment. If
       not set, it will use the training environment for evaluation too.

    """

    def __init__(
        self,
        algorithm_config: AlgorithmConfig,
        model_config: ModelConfig,
        seed: int,
        config: ExperimentConfig,
        env: MultiAgentNavgroundEnv | None = None,
        eval_env: MultiAgentNavgroundEnv | None = None,
        task: NavgroundTaskClass | None = None,
        critic_model_config: ModelConfig | None = None,
        callbacks: list[Callback] | None = None,
    ):
        if task is None and env is None:
            raise ValueError("At least one of env and task should be defined")
        if not task and env:
            task = NavgroundTaskClass('navground', {},
                                      env=env,
                                      eval_env=eval_env)
        super().__init__(task=task,
                         algorithm_config=algorithm_config,
                         model_config=model_config,
                         seed=seed,
                         config=config,
                         critic_model_config=critic_model_config,
                         callbacks=callbacks)

    def export_policy(self,
                      path: PathLike | None = None,
                      name: str = 'policy',
                      group: str = 'agent',
                      dynamic_batch_size: bool = True) -> Path:
        """
        Export the policy to onnx using :py:func:`navground.learning.onnx.export`.

        :param      path:                Where to save the policy. If not set,
                                         it save it in the experiment log folder
                                         under <name>_<group>.onnx.
        :param      name:                The name file (used as alternative to ``path``)
        :param      group:               The group name
        :param      dynamic_batch_size:  Whether to expose a dynamic batch size

        :returns:   The path of the saved file.
        """
        from ...onnx.export import export

        if path is None:
            path = Path(self.folder_name) / f"{name}_{group}.onnx"
        export(self.get_single_agent_policy(group),
               path=path,
               dynamic_batch_size=dynamic_batch_size)
        return Path(path)

    def export_policies(self,
                        path: PathLike | None = None,
                        name: str = 'policy',
                        dynamic_batch_size: bool = True) -> list[Path]:
        """
        Export the policy to onnx using :py:func:`navground.learning.onnx.export`.

        :param      path:                The directoru where to save the policies. If not set,
                                         it uses the experiment log folder.
        :param      name:                The files prefix
        :param      dynamic_batch_size:  Whether to expose a dynamic batch size

        :returns:   The path of the saved files.
        """
        from ...onnx.export import export

        if path is None:
            path = Path(self.folder_name)
        else:
            path = Path(path)
        paths = []
        for group in self.group_map:
            group_path = path / f"{name}_{group}.onnx"
            export(self.get_single_agent_policy(group),
                   path=group_path,
                   dynamic_batch_size=dynamic_batch_size)
            paths.append(group_path)
        return paths

    def load_policy(self,
                    path: PathLike | None = None,
                    name: str = 'policy',
                    group: str = 'agent') -> OnnxPolicy:
        """
        Loads an onnx policy.

        :param      path:  The path.
            If not set, it load <name>_<group>.onnx from the experiment log folder.

        :param      name:  The name file (used as alternative to ``path``)
        :param      group: The name of the group using the policy.
                           If provided it is used to associate an action space
                           to the policy.

        :returns:   The loaded policy.
        """
        from ...onnx.policy import OnnxPolicy

        if path is None:
            path = Path(self.folder_name) / f"{name}_{group}.onnx"

        if group:
            action_space = self.action_space(group)
        else:
            action_space = None
        return OnnxPolicy(path=path, action_space=action_space)

    def get_indices(self, group: str) -> list[int]:
        """
        Gets the indices of agents with name <group>_<index>

        :param      group:  The group

        :returns:   The agent indices
        """
        env = self.test_env._env
        return env.get_indices(group)

    def load_policies(self,
                      path: PathLike | None = None,
                      name: str = 'policy') -> list[GroupConfig]:
        """
        Loads the onnx policy for each group.

        :param      path:  The directory path.
            If not set, it uses the experiment log folder.

        :param      name:  The file prefix

        :returns:   A list of configurations with loaded policies
                    and agent indices set
        """
        return [
            GroupConfig(indices=Indices(self.get_indices(group)),
                        policy=self.load_policy(name="best_policy",
                                                group=group))
            for group in self.group_map
        ]

    def action_space(self, group: str) -> gym.Space:
        """
        Gets the action space associated with a group

        :param       group:  The group name

        :raises ValueError:  If the group is empty

        :returns:            The action space.
        """
        agents = self.group_map.get(group, [])
        if not agents:
            raise ValueError(f"Unknown group {group}!")
        env = self.test_env._env
        return env.action_space(agents[0])

    def observation_space(self, group: str) -> gym.Space:
        """
        Gets the observation space associated with a group

        :param       group:  The group name

        :raises ValueError:  If the group is empty

        :returns:            The action space.
        """
        agents = self.group_map.get(group, [])
        if not agents:
            raise ValueError(f"Unknown group {group}!")
        env = self.test_env._env
        return env.observation_space(agents[0])

    def get_single_agent_policy(self,
                                group: str = 'agent') -> SingleAgentPolicy:
        """
        Gets the policy associated with a group, as single agent
        (i.e., non-tensordict) policy.

        :param       group:  The group name

        :raises ValueError:  If the group is empty

        :returns:            The policy.
        """
        agents = self.group_map.get(group, [])
        if not agents:
            raise ValueError(f"Unknown group {group}!")
        env = self.test_env._env
        action_space = env.action_space(agents[0])
        observation_space = env.observation_space(agents[0])
        return SingleAgentPolicy(action_space=action_space,
                                 observation_space=observation_space,
                                 policy=self.group_policies[group])

    def get_single_agent_policies(self) -> list[GroupConfig]:
        """
        Gets the policy associated for each group, single agent
        (i.e., non-tensordict) policies.

        :returns:   A list of configurations with loaded policies
                    and agent indices set
        """
        env = self.test_env._env
        configs = []
        for group, agents in self.group_map.items():
            agent = agents[0]
            action_space = env.action_space(agent)
            observation_space = env.observation_space(agent)
            policy = SingleAgentPolicy(action_space=action_space,
                                       observation_space=observation_space,
                                       policy=self.group_policies[group])
            configs.append(
                GroupConfig(policy=policy,
                            indices=Indices(self.get_indices(group))))
        return configs

    def run_for(self, iterations: int = 0, steps: int = 0):
        """
        Train the policy for some iterations and/or steps.

        :param      iterations:  The iterations
        :param      steps:       The steps
        """
        n = self.config.collected_frames_per_batch(self.on_policy)
        target_iterations = self.n_iters_performed + max(
            iterations, int(math.ceil(steps / n)))
        target_steps = self.total_frames + max(steps, iterations * n)
        self.config.max_n_iters = max(target_iterations,
                                      self.config.max_n_iters)
        self.config.max_n_frames = max(target_steps, self.config.max_n_frames)
        self._setup_collector()
        self.run()

    @property
    def log_directory(self) -> Path:
        """
        The logging directory.

        :returns: The directory path
        """
        path = Path(self.folder_name)
        name = path.parts[-1]
        return path / name

    def load_log(self) -> pd.DataFrame:
        """
        Loads data stored in the the logs csv files.

        :returns:   A dataframe for all data.
        """
        import pandas as pd

        rs = []
        for path in (self.log_directory / 'scalars').glob('*.csv'):
            ds = pd.read_csv(path, index_col=[0], names=[str(path.stem)])
            rs.append(ds)
        return pd.concat(rs, axis=1)

    def evaluate_policy(
        self,
        n_eval_episodes: int = 10,
        return_episode_rewards: bool = False
    ) -> tuple[float, float] | tuple[list[float], list[int]]:
        """
        Evaluates the current policy using
        :py:func:navground.learning.utils.benchmarl.evaluate_policy`.

        :param      n_eval_episodes:         The number of episodes
        :param      return_episode_rewards:  Whether to return individual episode rewards
                                             (vs aggregate them)

        :returns:   If ``return_episode_rewards`` is set,
                    a tuple (list of cumulated episodes rewards', list of episodes length)
                    else, a tuple (average episodes rewards, std dev of episodes rewards)

        """
        return evaluate_policy(self.policy,
                               self.test_env,
                               n_eval_episodes=n_eval_episodes,
                               return_episode_rewards=return_episode_rewards)

    def plot_eval_logs(self,
                       group: str = 'agent',
                       two_axis: bool = False,
                       reward: bool = True,
                       reward_color: str = 'k',
                       reward_linestyle: str = '--',
                       success: bool = False,
                       length: bool = False,
                       reward_low: float | None = None,
                       reward_high: float = 0,
                       length_low: float = 0,
                       length_high: float | None = None,
                       **kwargs: Any) -> None:
        """
        Plots reward, success and/or length from the eval logs

        :param      group:    The name of the group
        :param      reward:       Whether to plot the reward
        :param      reward_color: The reward color
        :param      reward_linestyle: The reward linestyle
        :param      success:      Whether to plot the success
        :param      length:       Whether to plot the length
        :param      reward_low:   An optional lower bound to scale the reward
        :param      reward_high:  An optional upper bound to scale the reward
        :param      lenght_low:   An optional lower bound to scale the reward
        :param      lenght_high:  An optional upper bound to scale the reward
        :param      two_axis:     Whether to use two axis (only if there are two fields)
        """
        from ..plot import LogField, plot_logs

        fields: list[LogField] = []
        if reward:
            fields.append(
                LogField(label="mean reward",
                         key=f"eval_{group}_reward_episode_reward_mean",
                         linestyle=reward_linestyle,
                         color=reward_color,
                         low=reward_low,
                         high=reward_high))
        if success:
            fields.append(
                LogField(label="success rate",
                         key=f"eval_{group}_success_rate",
                         linestyle='-',
                         color='g'))
        if length:
            fields.append(
                LogField(label="mean length",
                         key="eval_reward_episode_len_mean",
                         linestyle=':',
                         color='b',
                         low=length_low,
                         high=length_high))
        plot_logs(self.load_log(),
                  two_axis=two_axis,
                  fields=fields,
                  key='counters_total_frames',
                  **kwargs)

    # Same as Experiment.reload_from_file but loading a NavgroundExperiment.
    @staticmethod
    def reload_from_file(restore_file: str) -> NavgroundExperiment:
        """
        Restores the experiment from the checkpoint file.

        This method expects the same folder structure created when an experiment is run.
        The checkpoint file (``restore_file``) is in the checkpoints directory
        and a config.pkl file is
        present a level above at ``restore_file/../../config.pkl``

        :param restore_file: The checkpoint file (.pt) of the experiment reload.

        :returns: The reloaded experiment.

        """
        import os
        import pickle

        experiment_folder = Path(restore_file).parent.parent.resolve()
        config_file = experiment_folder / "config.pkl"
        if not os.path.exists(config_file):
            raise ValueError("config.pkl file not found in experiment folder.")
        with open(config_file, "rb") as f:
            task = pickle.load(f)
            task_config = pickle.load(f)
            algorithm_config = pickle.load(f)
            model_config = pickle.load(f)
            seed = pickle.load(f)
            experiment_config = pickle.load(f)
            critic_model_config = pickle.load(f)
            callbacks = pickle.load(f)
        task.config = task_config
        experiment_config.restore_file = restore_file
        experiment = NavgroundExperiment(
            task=task,
            algorithm_config=algorithm_config,
            model_config=model_config,
            seed=seed,
            config=experiment_config,
            callbacks=callbacks,
            critic_model_config=critic_model_config,
        )
        # print(f"\nReloaded experiment {experiment.name} from {restore_file}.")
        return experiment



from __future__ import annotations

import dataclasses as dc
from collections.abc import Mapping
from typing import Any, cast

try:
    from typing import Self
except ImportError:
    try:
        from typing_extensions import Self
    except ImportError:
        ...

import numpy as np
from navground import core, sim

from .types import Reward


class NullReward(Reward, register_name="Zero"):
    """
        A dummy reward that returns always zero
    """

    def __call__(self, agent: sim.Agent, world: sim.World,
                 time_step: float) -> float:
        """
        A dummy zero reward.

        :param      agent:      The agent
        :param      world:      The world
        :param      time_step:  The time step

        :returns:   zero
        """
        return 0.0


class FixedReward(Reward, register_name="Fixed"):
    """
        A const -1 reward to favor shorter runs.
    """

    def __call__(self, agent: sim.Agent, world: sim.World,
                 time_step: float) -> float:
        """
        A const -1 reward.

        :param      agent:      The agent
        :param      world:      The world
        :param      time_step:  The time step

        :returns:   -1
        """
        return -1.0


class EfficacyReward(Reward, register_name="Efficacy"):
    """
        A reward that returns the agent's
        :py:attr:`navground.core.Behavior.efficacy`

    """

    def __call__(self, agent: sim.Agent, world: sim.World,
                 time_step: float) -> float:
        if agent.behavior:
            return min(agent.behavior.efficacy, 1) - 1
        return 0


class TargetEfficacyReward(Reward, register_name="TargetEfficacy"):
    """
        A reward that sums the linear *and* angular efficacy and
        penalize linear *and* angular speed once the agent has arrived.
    """

    def __call__(self, agent: sim.Agent, world: sim.World,
                 time_step: float) -> float:
        r: float = 0
        b = agent.behavior
        if b:
            if b.get_target_direction() is not None:
                r += min(b.efficacy, 1) - 1
                if b.kinematics.dof() < 3:
                    return r
            else:
                r -= abs(b.speed) / b.optimal_speed
            d = b.get_target_angular_velocity()
            if d:
                r += min(b.angular_speed / d, 1) - 1
            else:
                r -= abs(b.angular_speed) / b.optimal_angular_speed
        return r


class TargetDistanceReward(Reward, register_name="TargetDistance"):
    """
        A reward that penalizes target distances

    """

    def __call__(self, agent: sim.Agent, world: sim.World,
                 time_step: float) -> float:
        if agent.behavior:
            return -((agent.behavior.get_target_distance() or 0) +
                     (agent.behavior.get_target_angular_distance() or 0))
        return 0


@dc.dataclass
class SocialReward(Reward, register_name="Social"):
    """
    Reward function for social navigation, inspired by [TODO add citation]

    It returns a weighted sum of

    - violations of the social margin ([-1, 0], weight ``alpha``)
    - violations of the safety margin ([-1, 0], weight ``beta``)
    - efficacy ([-1, 0], weight 1)

    so that it is lower or equal to zero, which corresponds to no violations,
    while moving at optimal speed towards the target.

    :param      alpha:                  The weight of social margin violations
    :param      beta:                   The weight of safety violations
    :param      critical_safety_margin: Violation of this margin has maximal penalty of -1
    :param      safety_margin:          Violations between this and the critical
                                        safety_margin have a linear penalty. If not set,
                                        it defaults to the agent's own safety_margin.
    :param      beta:                   The weight of safety violation
    :param      default_social_margin:  The default social margin
    :param      social_margins:         The social margins assigned to neighbors' ids

    :returns:   A function that returns -1 if the safety margin is violated
                or weighted sum of social margin violations and efficacy.
    """
    alpha: float = 0.0
    """The weight of social margin violations"""
    beta: float = 1.0
    """The weight of safety margin violations"""
    critical_safety_margin: float = 0.0
    """
       The value above which we assign a maximal penalty
       to safety margin violations
    """
    safety_margin: float | None = None
    """An optional value. If not set, it will use the behavior's safety margin"""
    default_social_margin: float = 0.0
    """The default value of the social margin"""
    social_margins: dict[int, float] = dc.field(default_factory=dict)
    """The social margins to be applied to specific type of neighbors"""

    def __post_init__(self) -> None:
        self._social_margin = core.SocialMargin(self.default_social_margin)
        for i, m in self.social_margins.items():
            self._social_margin.set(m, i)
        self._max_social_margin = self._social_margin.max_value

    def __call__(self, agent: sim.Agent, world: sim.World,
                 time_step: float) -> float:
        if self.safety_margin is None:
            if agent.behavior:
                sm = agent.behavior.safety_margin
            else:
                sm = 0
        else:
            sm = self.safety_margin
        if sm > 0:
            max_violation = max(0, sm - self.critical_safety_margin)
            sv = world.compute_safety_violation(agent, sm)
            # if sv >= sm:
            #     return -1.0
            if sv == 0:
                r = 0.0
            elif sv > max_violation:
                r = -self.beta
            else:
                r = -self.beta * sv / max_violation
        else:
            r = 0
        if self._max_social_margin > 0 and self.alpha != 0:
            ns = world.get_neighbors(agent, self._max_social_margin)
            for n in ns:
                distance = cast("float",
                                np.linalg.norm(n.position - agent.position))
                margin = self._social_margin.get(n.id, distance)
                if margin > distance:
                    r += (distance - margin) * self.alpha * time_step
        if agent.task and agent.task.done():
            r += 1.0
        if agent.behavior:
            r += np.clip(0, 1, agent.behavior.efficacy) - 1
        return r

    def _get_dict(self) -> dict[str, Any]:
        return dc.asdict(self)

    @classmethod
    def _make_from_dict(cls, value: Mapping[str, Any]) -> Self:
        return cls(**value)



from __future__ import annotations

import dataclasses as dc

import gymnasium as gym
import numpy as np
from navground import core

from ..types import Action
from .control_action import ControlActionConfig


@dc.dataclass
class ControlActionWithCommConfig(ControlActionConfig,
                                  register_name="ControlWithComm"):
    """
    Configuration of the conversion between actions
    and control commands.

    Actions are composed of control values (accelerations
    or command velocities, depending on
    :py:meth:`ControlActionConfig.use_acceleration_action`)
    and the message to be transmitted.

    Action are decomposed as:

    - the control is converted to a :py:class:`navground.core.Twist2` (and later actuated)
    - the message is set as attribute ``Behavior._comm`` (and later broadcasted)

    :param dtype: The data type

    :param dof: The number of degrees of freedom of the agent

    :param max_speed: The upper bound of the speed.

    :param max_angular_speed: The upper bound of the angular speed.

    :param max_acceleration: The upper bound of the acceleration.

    :param max_angular_acceleration: The upper bound of the angular acceleration.

    :param use_acceleration_action: Whether actions are accelerations.
                                    If not set, actions are velocities.

    :param use_wheels: Whether action uses wheel speeds/acceleration
                       instead of body speeds/acceleration.
                       Only effective if the b behavior has a wheeled kinematics.

    :param fix_orientation: Whether to force the agent not to control orientation,
                            i.e., to not include the angular command in actions.

    :param has_wheels: Whether the agent as wheels.
                       If None, it will defer to the agent kinematics.

    :param comm_size: The size of the message to broadcast
    """
    comm_size: int = 2
    """The size of the message to broadcast"""

    @property
    def action_size(self) -> int:
        return super().action_size + self.comm_size

    @property
    def comm_space(self) -> gym.spaces.Box:
        """
        The box space of the broadcasted messages with dimension
        :py:meth:`ControlActionConfig.action_size`
        and values between -1 and 1.

        :returns: the box space
        """
        return gym.spaces.Box(low=-1,
                              high=1,
                              shape=(self.comm_size, ),
                              dtype=core.FloatType)

    def get_cmd_from_action(self, action: Action,
                            behavior: core.Behavior | None,
                            time_step: float) -> core.Twist2:
        cmd = super().get_cmd_from_action(action, behavior, time_step)
        if behavior:
            behavior._comm = action[  # type: ignore[attr-defined]
                -self.comm_size:]
        return cmd

    def get_action(self, behavior: core.Behavior, time_step: float) -> Action:
        act = ControlActionConfig.get_action(self, behavior, time_step)
        if hasattr(behavior, '_comm'):
            comm = behavior._comm  # type: ignore[attr-defined]
        else:
            comm = np.zeros(self.comm_size, dtype=core.FloatType)
        act = np.concatenate([act, comm], dtype=core.FloatType)
        return act



from __future__ import annotations

import warnings
from collections.abc import Callable
from typing import TYPE_CHECKING, Any

import numpy as np

from ..env import BaseEnv
from ..types import AnyPolicyPredictor, accept_info

if TYPE_CHECKING:
    from stable_baselines3.common.vec_env import VecEnv


def evaluate_policy(
    model: AnyPolicyPredictor,
    env: BaseEnv | VecEnv,
    n_eval_episodes: int = 10,
    deterministic: bool = True,
    render: bool = False,
    callback: Callable[[dict[str, Any], dict[str, Any]], None] | None = None,
    reward_threshold: float | None = None,
    return_episode_rewards: bool = False,
    warn: bool = True,
) -> tuple[float, float] | tuple[list[float], list[int]]:
    """
    Extends StableBaseline3 :py:func:`stable_baselines3.common.evaluation.evaluate_policy`
    to also accept :py:class:`navground.learning.types.PolicyPredictorWithInfo` models.
    """
    from stable_baselines3.common.vec_env import (DummyVecEnv, VecEnv,
                                                  VecMonitor,
                                                  is_vecenv_wrapped)

    is_monitor_wrapped = False
    # Avoid circular import
    from stable_baselines3.common.monitor import Monitor

    if not isinstance(env, VecEnv):
        env = DummyVecEnv([
            lambda: env
        ])

    is_monitor_wrapped = is_vecenv_wrapped(
        env, VecMonitor) or env.env_is_wrapped(Monitor)[0]

    if not is_monitor_wrapped and warn:
        warnings.warn(
            "Evaluation environment is not wrapped with a ``Monitor`` wrapper. "
            "This may result in reporting modified episode lengths and rewards, "
            "if other wrappers happen to modify these. "
            "Consider wrapping environment first with ``Monitor`` wrapper.",
            UserWarning,
            stacklevel=1,
        )

    n_envs = env.num_envs
    episode_rewards = []
    episode_lengths = []

    episode_counts = np.zeros(n_envs, dtype="int")
    # Divides episodes among different sub environments in the vector as evenly as possible
    episode_count_targets = np.array([(n_eval_episodes + i) // n_envs
                                      for i in range(n_envs)],
                                     dtype="int")

    current_rewards = np.zeros(n_envs)
    current_lengths = np.zeros(n_envs, dtype="int")
    observations = env.reset()
    infos = env.reset_infos
    states = None
    episode_starts = np.ones((env.num_envs, ), dtype=bool)
    pass_info = accept_info(model.predict)
    while (episode_counts < episode_count_targets).any():
        if pass_info:
            kwargs = {'info': infos}
        else:
            kwargs = {}
        actions, states = model.predict(
            observations,  # type: ignore[arg-type]
            state=states,
            episode_start=episode_starts,
            deterministic=deterministic,
            **kwargs)
        new_observations, rewards, dones, infos = env.step(actions)
        current_rewards += rewards
        current_lengths += 1
        for i in range(n_envs):
            if episode_counts[i] < episode_count_targets[i]:
                # unpack values so that the callback can access the local variables
                reward = rewards[i]
                done = dones[i]
                info = infos[i]
                episode_starts[i] = done

                if callback is not None:
                    callback(locals(), globals())

                if dones[i]:
                    if is_monitor_wrapped:
                        # Atari wrapper can send a "done" signal when
                        # the agent loses a life, but it does not correspond
                        # to the true end of episode
                        if "episode" in info.keys():  # noqa
                            # Do not trust "done" with episode endings.
                            # Monitor wrapper includes "episode" key in info if environment
                            # has been wrapped with it. Use those rewards instead.
                            episode_rewards.append(info["episode"]["r"])
                            episode_lengths.append(info["episode"]["l"])
                            # Only increment at the real end of an episode
                            episode_counts[i] += 1
                    else:
                        episode_rewards.append(current_rewards[i])
                        episode_lengths.append(current_lengths[i])
                        episode_counts[i] += 1
                    current_rewards[i] = 0
                    current_lengths[i] = 0

        observations = new_observations

        if render:
            env.render()

    mean_reward = np.mean(episode_rewards)
    std_reward = np.std(episode_rewards)
    if reward_threshold is not None:
        assert mean_reward > reward_threshold, (
            "Mean reward below threshold: "
            f"{mean_reward:.2f} < {reward_threshold:.2f}")
    if return_episode_rewards:
        return episode_rewards, episode_lengths
    return mean_reward, std_reward



from __future__ import annotations

import functools
from collections.abc import Callable, Collection
from typing import TYPE_CHECKING, cast

from navground import sim

from ..config import GroupConfig
from ..env import BaseEnv
from ..internal.base_env import NavgroundBaseEnv
from ..internal.group import create_agents_in_groups
from ..parallel_env import BaseParallelEnv

if TYPE_CHECKING:
    from imitation.data.types import DictObs


# TODO(Jerome): check that it works with flat observations
class GymProbe(sim.Probe):
    """
    A probe to record observation, rewards and actions, like
    during a rollout. Internally uses a
    :py:class:`imitation.data.rollout.TrajectoryAccumulator`
    to store the data, which it writes to datasets
    only at the end of the run:

    - observations/<agent_index>/<key>
    - actions/<agent_index>
    - rewards/<agent_index>

    :param      groups:  The configuration of the groups

    """

    def __init__(self, groups: Collection[GroupConfig]):
        super().__init__()
        self._groups = groups

    def prepare(self, run: sim.ExperimentalRun) -> None:
        from imitation.data.rollout import TrajectoryAccumulator

        self._agents = create_agents_in_groups(run.world, self._groups)
        self._tas = {
            k: TrajectoryAccumulator()  # type: ignore[no-untyped-call]
            for k in self._agents
        }
        self._first = True
        if self._first:
            self._add_obs(run.world)
            self._first = False

    def update(self, run: sim.ExperimentalRun) -> None:
        from imitation.data.types import maybe_wrap_in_dictobs

        for i, agent in self._agents.items():
            agent.update_state(run.world)
            if agent.gym:
                obs = agent.gym.update_observation()
                acts = agent.gym.get_action(run.time_step)
                self._tas[i].add_step({
                    'obs': maybe_wrap_in_dictobs(obs),
                    'acts': acts,
                    'infos': {},
                    'rews': 0.0  # type: ignore
                })

    def finalize(self, run: sim.ExperimentalRun) -> None:

        for agent_index, ta in self._tas.items():
            ts = ta.finish_trajectory(None, False)
            for key, data in cast(
                    'DictObs',
                    ts.obs).items():  # type: ignore[no-untyped-call]
                run.add_record(f"observations/{agent_index}/{key}", data)
            run.add_record(f"actions/{agent_index}", ts.acts)
            run.add_record(f"rewards/{agent_index}", ts.rews)

    def _add_obs(self, world: sim.World) -> None:
        from imitation.data.types import maybe_wrap_in_dictobs

        world._prepare()
        for i, agent in self._agents.items():
            agent.update_state(world)
            if agent.gym:
                obs = agent.gym.update_observation()
                self._tas[i].add_step({'obs': maybe_wrap_in_dictobs(obs)})

    @classmethod
    def with_env(cls,
                 env: BaseEnv | BaseParallelEnv) -> Callable[[], GymProbe]:
        """
        Creates a probe factory to record all actions and observations
        in an environment

        :param      env:  The environment

        :returns:   A callable that can be added to runs or experiments,
            using :py:meth:`navground.sim.ExperimentalRun.add_record_probe` or
            :py:meth:`navground.sim.Experiment.add_record_probe`
        """
        assert isinstance(env.unwrapped, NavgroundBaseEnv)
        return functools.partial(cls, groups=env.unwrapped.groups_config)



from __future__ import annotations

import dataclasses as dc
from typing import Any, cast

import gymnasium as gym
import numpy as np
from navground import sim

from ...config import ControlActionConfig, ObservationConfig, StateConfig
from ...env import BaseEnv
from ...parallel_env import BaseParallelEnv, shared_parallel_env
from ...rewards import EfficacyReward
from ...scenarios.pad import PadScenario, are_two_agents_on_the_pad
from ...state_estimations import CommSensor
from ...types import Reward, SensorSequenceLike


@dc.dataclass
class PadReward(EfficacyReward, register_name="Pad"):
    """
    An efficacy reward that also penalizes by ``pad_penalty`` when both
    agents are inside the pad area.

    When ``neighbor_weight > 0``, it includes the efficacy of the neighbor,
    weighted accordingly.
    """
    pad_penalty: float = 10
    neighbor_weight: float = 0

    def __call__(self, agent: sim.Agent, world: sim.World,
                 time_step: float) -> float:
        r = super().__call__(agent, world, time_step)
        if are_two_agents_on_the_pad(world, 0):
            r -= self.pad_penalty
        if self.neighbor_weight:
            r += self.neighbor_weight * sum(
                EfficacyReward.__call__(self, other, world, time_step)
                for other in world.agents if other is not agent)
        return r


def is_success(agent: sim.Agent, world: sim.World) -> bool:
    if agent.behavior and agent.behavior.target.direction is not None:
        return bool(
            np.dot(agent.position, agent.behavior.target.direction) >
            world.bounding_box.max_x)
    return False


def is_failure(agent: sim.Agent, world: sim.World) -> bool:
    return are_two_agents_on_the_pad(world)


def neighbor(
        range: float = 10,
        max_speed: float = 0.166
) -> sim.state_estimations.DiscsStateEstimation:
    """
    The sensors that detects the neighbor.

    :param      range:      The range
    :param      max_speed:  The neighbor maximum speed

    :returns:   The sensor
    """
    return sim.state_estimations.DiscsStateEstimation(number=1,
                                                      range=range,
                                                      max_speed=max_speed,
                                                      name="neighbor",
                                                      include_valid=False,
                                                      include_y=False,
                                                      use_nearest_point=False)


def marker(min_x: float = -1,
           max_x: float = 1) -> sim.state_estimations.MarkerStateEstimation:
    """
    The sensors that detects the pad.

    :param      min_x:  The lower bound of the relative (horizontal) position.
    :param      max_x:  The upper bound of the relative (horizontal) position.

    :returns:   The sensor
    """
    return sim.state_estimations.MarkerStateEstimation(
        reference_orientation=sim.state_estimations.MarkerStateEstimation.
        ReferenceOrientation.target_direction,
        name="pad",
        min_x=min_x,
        max_x=max_x,
        include_y=False)


def comm(size: int = 1,
         name: str = 'neighbor',
         binarize: bool = False) -> sim.Sensor:
    """
    The sensors that receives messages broadcasted by the neighbor.

    :param      size:      The size of the message
    :param      name:      The namespace
    :param      binarize:  Whether to binarize the message

    :returns:   The sensor
    """
    return CommSensor(binarize=binarize, size=size, name=name)


def get_env(action: ControlActionConfig,
            observation: ObservationConfig,
            sensors: SensorSequenceLike = tuple(),
            reward: Reward = PadReward(),
            max_duration: float = 20,
            time_step: float = 0.1,
            init_success: bool = False,
            intermediate_success: bool = False,
            include_success: bool = True,
            render_mode: str | None = None,
            render_kwargs: dict = {},
            state: StateConfig | None = None,
            multi_agent: bool = True,
            **kwargs: Any) -> BaseEnv | BaseParallelEnv:
    """
    Creates the an environment where 2 agents cross along a corridor where
    there is pad which should not be entered by more than one agent at the
    same time.

    :param      action:                The action config
    :param      observation:           The observation config
    :param      sensors:               The sensors
    :param      reward:                The reward function
    :param      max_duration:          The maximal duration [s]
    :param      time_step:             The time step [s]
    :param      init_success:          The initialization value for intermediate success
    :param      intermediate_success:  Whether to return intermediate success
    :param      include_success:       Whether to include success
    :param      render_mode:           The render mode
    :param      render_kwargs:         The rendering keywords arguments
    :param      state:                 The global state config
                                       (only relevant if ``multi_agent=True``)
    :param      multi_agent:           Whether the environments controls both agents
    :param      kwargs:                Keywords arguments passed to
                                       :py:class:`navground.learning.scenarios.PadScenario`.

    :returns: A Parallel PettingZoo environment if `multi_agent` is set,
        else a Gymnasium environment.
    """

    if not multi_agent:
        return gym.make('navground',
                        scenario=PadScenario(**kwargs),
                        action=action,
                        observation=observation,
                        reward=reward,
                        sensors=sensors,
                        time_step=time_step,
                        success_condition=is_success,
                        failure_condition=is_failure,
                        terminate_on_success=True,
                        terminate_on_failure=False,
                        init_success=init_success,
                        intermediate_success=intermediate_success,
                        include_success=include_success,
                        max_duration=max_duration)
    return shared_parallel_env(scenario=PadScenario(**kwargs),
                               action=action,
                               observation=observation,
                               reward=reward,
                               sensors=sensors,
                               time_step=time_step,
                               success_condition=is_success,
                               failure_condition=is_failure,
                               terminate_on_success=True,
                               terminate_on_failure=False,
                               max_duration=max_duration,
                               wait=True,
                               init_success=init_success,
                               include_success=include_success,
                               intermediate_success=intermediate_success,
                               render_mode=render_mode,
                               render_kwargs=render_kwargs,
                               state=state)


def plot_policy(policy: Any,
                cmap: str = 'RdYlGn',
                title='Distributed policy',
                speed: float = 0.01):
    from matplotlib import pyplot as plt

    xs = np.linspace(-1, 1, 101, dtype=np.float32)
    ys = np.linspace(-2, 2, 101, dtype=np.float32)
    xv, yv = np.meshgrid(xs, ys)
    if isinstance(policy.observation_space, gym.spaces.Dict):
        obs = {
            'pad/x': xv,
            'neighbor/position': yv,
            'neighbor/velocity': np.full(101 * 101, speed, dtype=np.float32),
            'ego_velocity': np.full(101 * 101, speed, dtype=np.float32)
        }
        obs = {
            k: obs[k].reshape(-1,
                              *cast('gym.spaces.Box', v).shape)
            for k, v in policy.observation_space.items()
        }
    else:
        raise NotImplementedError
    act, _ = policy.predict(obs, deterministic=True)
    act = act.reshape(101, 101, -1)
    if isinstance(policy.action_space, gym.spaces.Discrete):
        act = act - 1
    elif isinstance(policy.action_space, gym.spaces.MultiBinary):
        act = act[..., 1] - act[..., 0]
    plt.figure(figsize=(6, 4))
    plt.imshow(act, vmin=-1, vmax=1, cmap=cmap)
    plt.xlabel('relative pad center position')
    plt.ylabel('neighbor position')
    plt.colorbar()
    plt.xticks(np.linspace(0, 101, 5),
               [f'{x:.2f}' for x in np.linspace(-1, 1, 5).round(2)])
    plt.yticks(np.linspace(0, 101, 5),
               [f'{x:.2f}' for x in np.linspace(-2, 2, 5).round(2)])
    plt.title(title)


def plot_policy_with_comm(policy: Any,
                          cmap: str = 'RdYlGn',
                          title='Distributed policy with comm',
                          speed: float = 0.01,
                          binarize: bool | None = None):
    from matplotlib import pyplot as plt

    xs = np.linspace(-1, 1, 101, dtype=np.float32)
    if isinstance(policy.observation_space, gym.spaces.Dict):
        if binarize is None:
            comm_space = policy.observation_space['neighbor/comm']
            binarize = not np.issubdtype(comm_space.dtype, np.floating)
        low = -1
        high = 1
        # low = comm_space.low[0]
        # high = comm_space.high[0]
        if not binarize:
            ys = np.linspace(low, high, 101, dtype=np.float32)
            xv, yv = np.meshgrid(xs, ys)
            obs = {
                'pad/x': xv,
                'neighbor/comm': yv,
                'ego_velocity': np.full(101 * 101, speed, dtype=np.float32)
            }
            obs = {
                k: obs[k].reshape(-1,
                                  *cast('gym.spaces.Box', v).shape)
                for k, v in policy.observation_space.items()
            }
            act, _ = policy.predict(obs, deterministic=True)
            act = act.reshape(101, 101, -1)
            # if isinstance(policy.action_space, gym.spaces.Discrete):
            #     act = act - 1
            # elif isinstance(policy.action_space, gym.spaces.MultiBinary):
            #     act = act[..., 1] - act[..., 0]
            fig, axs = plt.subplots(ncols=2, figsize=(12, 4))
            for i, (ax, t) in enumerate(
                    zip(axs, ('acceleration', 'tx'), strict=True)):
                im = ax.imshow(act[..., i], vmin=-1, vmax=1, cmap=cmap)
                ax.set_xlabel('relative pad center position')
                ax.set_ylabel('rx')
                ax.set_xticks(np.linspace(0, 101, 5),
                              np.linspace(-1, 1, 5).round(2))
                ax.set_yticks(np.linspace(0, 101, 5),
                              np.linspace(-1, 1, 5).round(2))
                ax.title.set_text(t)
            fig.colorbar(im, ax=axs.ravel().tolist())
            fig.suptitle(title)
        else:
            xv = np.concatenate([xs, xs], axis=-1)
            yv = np.array([0] * len(xs) + [1] * len(xs), dtype=np.float32)
            obs = {
                'pad/x': xv,
                'neighbor/comm': yv,
                'ego_velocity': np.full(len(xv), speed, dtype=np.float32)
            }
            obs = {
                k: obs[k].reshape(-1,
                                  *cast('gym.spaces.Box', v).shape)
                for k, v in policy.observation_space.items()
            }
            act, _ = policy.predict(obs, deterministic=True)
            if isinstance(policy.action_space, gym.spaces.Discrete):
                act = np.stack([(act % 3) - 1, act // 3], axis=-1)
            elif isinstance(policy.action_space, gym.spaces.MultiBinary):
                raise NotImplementedError
            act = act.reshape(2, 101, 2)
            fig, axs = plt.subplots(ncols=2, figsize=(12, 4))
            for i, (ax, t) in enumerate(
                    zip(axs, ('acceleration', 'tx'), strict=True)):
                if i == 1:
                    ax.plot(xs, (act[0, ..., i] > 0).astype(int), label="RX=0")
                    ax.plot(xs, (act[1, ..., i] > 0).astype(int), label="RX=1")
                else:
                    ax.plot(xs, act[0, ..., i], label="RX=0")
                    ax.plot(xs, act[1, ..., i], label="RX=1")
                ax.set_xlabel('relative pad center position')
                ax.set_ylabel(t)
                ax.legend()
            fig.suptitle(title)
    else:
        raise NotImplementedError



from .base import make_logger, BaseILAlgorithm
from .bc import BC
from .dagger import DAgger
from .utils import make_vec_from_env
from .pz_utils import make_vec_from_penv
from .tqdm_utils import setup_tqdm

__all__ = [
    'BaseILAlgorithm', 'BC', 'DAgger', 'make_logger', 'make_vec_from_env',
    'make_vec_from_penv', 'setup_tqdm'
]
